#######################################################################
#                                                                     #
#                                 Pph                                 #
#                                                                     #
#                    The Portable Project Handler                     #
#                                                                     #
#               EPI Pomdapi, INRIA Paris - Rocquencourt               #
#                                                                     #
#  Copyright 2005-2013 INRIA.                                         #
#  Institut National de Recherche en Informatique et en Automatique.  #
#  All rights reserved. This file is distributed only by permission.  #
#                                                                     #
#  Francois Clement <Francois.Clement@inria.fr>                       #
#  Pierre Weis <Pierre.Weis@inria.fr>                                 #
#                                                                     #
#######################################################################

# $Id: script_functions.profile,v 1.11 2013-11-16 11:14:15 fclement Exp $

#
# Pph system script functions.
# Used by [un]configure[_user] and pph scripts.
#

SCRIPT=`basename $0`

#
# Error and warning messages.
#

# Local function.
gen_print_message () { $ECHO "$1"; }

# Local function.
print_tagged_message () { gen_print_message "$SCRIPT: $1."; }

# Local function.
print_final_tagged_message () {
    $ECHO ""
    print_tagged_message "$1"
    $ECHO ""
}

# Print a typed message on the standard error.
# Local function.
print_types_message_on_stderr () {
    ERROR_TYPE="$1"
    ERROR_MESSAGE="$2"
    
    print_final_tagged_message "$ERROR_TYPE: $ERROR_MESSAGE" 1>&2
}

# Print an error message.
error () { print_types_message_on_stderr "error" "$1"; }

# Print an internal error message.
internal_error () {
    print_types_message_on_stderr "internal error" "$1. Please report"
}

# Print a warning message.
warning () { print_types_message_on_stderr "warning" "$1"; }

# Print a remedy message.
remedy () {
    if test -n "$1"; then
        MESSAGE="$1"
    else
        MESSAGE="For help, run
	$SCRIPT --help
"
    fi
    gen_print_message "$MESSAGE" 1>&2
}

#
# Debug, report, see-you and normal operating modes.
#
# mode    | echo message | echo command | run command
# ---------------------------------------------------
# debug   | yes          | yes          | yes
# report  | yes          | no           | yes
# see_you | yes          | yes          | no
# normal  | no           | no           | yes
#

# Print a report message when appropriate on the standard output.
report_message () {
    MESSAGE="$1"
    
    # Don't use test boolean operators -o or -a, as they are not universal.
    # Use shell operators || or && instead.
    if ${DEBUG:?} || ${REPORT:?} || ${SEE_YOU:?}; then
        print_tagged_message "$MESSAGE"
    fi
}

# Report a message and append it to a log file.
report_and_log_message () {
    MESSAGE="$1"
    LOG_FILE="$2"
    
    report_message "$MESSAGE" | $TEE_APPEND "${LOG_FILE:?}"
}
echo_and_log_message () {
    MESSAGE="$1"
    LOG_FILE="$2"
    
    $ECHO "$MESSAGE" | $TEE_APPEND "${LOG_FILE:?}"
}

# Run and/or echo a command when appropriate.
run_command () {
    CMD="$1"
    
    # Don't use the test boolean operators -o or -a, as they don't seem to work
    # properly everywhere. Use shell operators || or && instead.
    if ${DEBUG:?} || ${SEE_YOU:?}; then
        $ECHO "$CMD"
    fi
    
    if ! ${SEE_YOU:?}; then
        eval "$CMD"
        IER=$?
        if test $IER -ne 0; then exit $IER; fi
    fi
}

# Run a command and hide its output.
run_command_hide_output () { run_command "$1" 1> /dev/null 2>&1; }

#
# Warning messages in automatically generated files.
#

# Unused.
#gen_newline () {
#    WHOS_FOR="$1"
#    
#    case "${WHOS_FOR:?}" in
#    ( for_sed )
#        NEWLINE="\n\\
#"
#        ;;
#    ( * )
#        NEWLINE="
#"
#        ;;
#    esac
#}

AUTO_MSG="Automatically generated by Pph."

# Unused.
#gen_dont_warning_msg_for_sed () {
#    SOURCE_FILE="$1"
#    
#    if test -n "$SOURCE_FILE"; then
#        RATHER_MSG=", but rather its source: \"$SOURCE_FILE\""
#    else
#        RATHER_MSG=""
#    fi
#    
#    MSG="# $AUTO_MSG\n\
##\n\
## Do not edit this file$RATHER_MSG."
#    
#    $ECHO "$MSG"
#}

# Local function.
gen_dont_warning_msg_for_echo () {
    SOURCE_FILE="$1"
    
    if test -n "$SOURCE_FILE"; then
        RATHER_MSG=", but rather its source: \"$SOURCE_FILE\""
    else
        RATHER_MSG=""
    fi
    
    MSG="# $AUTO_MSG
#
# Do not edit this file$RATHER_MSG.

"
    
    $ECHO "$MSG"
}

gen_dont_warning_msg_for_tex () {
    SOURCE_FILE="$1"
    
    if test -n "$SOURCE_FILE"; then
        RATHER_MSG=", but rather its source: \"$SOURCE_FILE\""
    else
        RATHER_MSG=""
    fi
    
    MSG="% $AUTO_MSG
%
% Do not edit this file$RATHER_MSG.

"
    
    $ECHO "$MSG"
}

# Local function.
gen_do_warning_msg_for_echo () {
    MSG="# $AUTO_MSG
#
# Feel free to modify the value of the variable(s), then run configure.
# Remember that each variable must be assigned only once.
# Note also that values must be enclosed between double quotes (\"),
# since we want to allow white spaces in values.
"
    
    $ECHO "$MSG"
}

#
# Shortcuts.
#

make_quiet () {
    $MAKE_COMMAND VERBOSITY_LEVEL="${VERBOSITY_LEVEL:?}" "$1" 1> /dev/null
}

#
# Files operations.
#

# Copy source file into target only when the target does not exist.
cautious_copy () {
    SOURCE="$1"
    TARGET="$2"
    
    if ! test -f "${TARGET:?}"; then
        $CP "${SOURCE:?}" "$TARGET"
    fi
}

# Get file type.
# We could use LF="ls -F" and a case statement.
get_file_type () {
    FILE="$1"
    
    if test -f "${FILE:?}"; then
        FILE_TYPE="regular_file"
    elif test -d "$FILE"; then
        FILE_TYPE="directory"
    elif test -h "$FILE"; then
        FILE_TYPE="symbolic_link"
    else
        FILE_TYPE="other"
    fi
    $ECHO "$FILE_TYPE"
}

# Print file contents.
print_file_contents () {
    FILE="$1"
    
    FILE_TYPE=`get_file_type "$FILE"`
    IER=$?
    if test $IER -ne 0; then exit $IER; fi
    case "${FILE_TYPE:?}" in
    ( regular_file ) $CAT "${FILE:?}";;
    ( directory ) $LS "${FILE:?}";;
    ( symbolic_link ) $ECHO " -> `$LL ${FILE:?} | $AWK '{print $NF}'`";;
    ( other ) $ECHO "  *** Cannot print info about \"$FILE\" ***";;
    esac
}

# Filter out the output of a quiet recursive diff.
grep_differ () {
    LIST="$1"
    SEP="$2"
    
    $ECHO "$LIST" | $GREP differ | \
    $AWK '{print $2}' | $AWK -F"${SEP:?}" '{print $2}'
}

grep_only () {
    LIST="$1"
    SEP="$2"
    
    $ECHO "$LIST" | $GREP "Only in $SEP" | \
    $SED -e 's@: @/@' | $AWK -F"${SEP:?}" '{print $2}'
}

filter_out_versioning_id_and_banner () {
    FILE1="$1"
    FILE2="$2"
    
    EMPTY_STRING=""
    
    $DIFF "${FILE1:?}" "${FILE2:?}" | \
    $EGREPV '^---$' | \
    $EGREPV "\\\$I${EMPTY_STRING}d: .* \\\$" | \
    $EGREPV '^> $' | \
    $EGREPV '^> #.*#$' | \
    $EGREPV '^> %.*%$' | \
    $EGREPV '^> <!--.*-->$' | \
    $SED -e 's/^\([134]\)c\1,[1-9][0-9]*$//' | \
    $SED -e 's/^\([1-9][0-9]*\)c\1$//'
}

print_list () {
    LIST="$1"
    
    MY_LIST=`$ECHO $LIST | $STRIP_SPACE`
    if test -n "$MY_LIST"; then
        $ECHO ""
        for ITEM in $LIST; do
            $ECHO "    $ITEM"
        done
    fi
}

#
# Strings.
#

capitalize () { $ECHO "\$(capitalize \"$1\")" | $HTMLC_FILTER; }
lowercase () { $ECHO "\$(lowercase \"$1\")" | $HTMLC_FILTER; }
uppercase () { $ECHO "\$(uppercase \"$1\")" | $HTMLC_FILTER; }
current_iso_date () { $ECHO '$(the_full_iso_date)' | $HTMLC_FILTER; }

# Local function.
strip_last_suffix () { $TR -d "$1"; }

#
# Variables.
#

# Return the new value only when the variable was previously empty.
single_assignment () {
    VARIABLE_NAME="$1"
    OLD_VALUE="$2"
    NEW_VALUE="$3"
    REMEDY_MESSAGE="$4"
    
    if test -z "$OLD_VALUE"; then
        $ECHO "$NEW_VALUE"
    else
        error \
            "$VARIABLE_NAME is already set to \"$OLD_VALUE\", \
cannot set it again to \"$NEW_VALUE\""
        remedy "$REMEDY_MESSAGE"
        exit ${ALREADY_SET_ERROR:?}
    fi
}

# Return the new value if the variable was previously empty.
default_assignment () {
    OLD_VALUE="$1"
    NEW_VALUE="$2"
    
    if test -z "$OLD_VALUE"; then
        $ECHO "$NEW_VALUE"
    else
        $ECHO "$OLD_VALUE"
    fi
}

# Unused.
#check_nonempty () {
#    VARIABLE_NAME="$1"
#    VALUE="$2"
#    
#    if test -z "$VALUE"; then
#        error "$VARIABLE_NAME must be nonempty"
#        exit ${EMPTY_VALUE_ERROR:?}
#    fi
#}

# Check that the value of a variable is a valid Pph file_name.
check_valid_file_name () {
    FILE_KIND="$1"
    FILE_PATH="$2"
    
    REJECT_WRONG_CHARS=`\
        $ECHO "$FILE_PATH" | \
        $AWK '/^[A-Za-z0-9_\.\-+]+$/{print $0}'`
    if test -z "$REJECT_WRONG_CHARS"; then
        error "invalid $FILE_KIND \"$FILE_PATH\".
Valid characters in Pph file_names are [A-Za-z0-9_\.\-+]"
        exit ${INVALID_ARGUMENT_ERROR:?}
    fi
}

# Check that a directory is nonexistant.
check_new_directory () {
    DIR="$1"
    
    if test -d "${DIR:?}"; then
        error "directory \"$DIR\" already exists"
        exit ${INVALID_VALUE_ERROR:?}
    fi
}

# Unused.
## Print items of a colon (':') separated list with a specified output
## separator.
#print_colon_list () {
#    OUTPUT_SEP="$1"
#    COLON_LIST="$2"
#    
#    $ECHO "$COLON_LIST" | \
#    $SED -e "s/:/$OUTPUT_SEP/g"
#}

# Unused.
#text_print_colon_list () { print_colon_list ", " "$1"; }

# Test if a value is in a  colon (':') separated list of values.
# Local function.
is_in_colon_list () {
    VALID_COLON_LIST="$1"
    VALUE="$2"
    
    case ":$VALID_COLON_LIST:" in
    ( *:$VALUE:* ) true;;
    ( * ) false;;
    esac
}

# Unused.
## Check that the nonempty value of a variable is in a  colon (':') separated
## list of valid nonempty values.
#check_in_colon_list () {
#    VALID_COLON_LIST="$1"
#    VARIABLE_NAME="$2"
#    VALUE="$3"
#    
#    if ! is_in_colon_list "${VALID_COLON_LIST:?}" "${VALUE:?}"; then
#        error "unknown $VARIABLE_NAME \"$VALUE\""
#        exit ${INVALID_VALUE_ERROR:?}
#    fi
#}

# Unused.
#validate () {
#    VALID_SET="$1"
#    VARIABLE_NAME="$2"
#    VALUE="$3"
#    
#    eval "
#    case \"${VALUE:?}\" in
#    ( ${VALID_SET:?} ) :;;
#    ( * )
#        error \"unknown $VARIABLE_NAME \\\"$VALUE\\\"\"
#        exit ${INVALID_VALUE_ERROR:?}
#        ;;
#    esac
#    "
#}

# Expand any variables in the expression.
expand () { eval '$ECHO "$1"'; }

# Expand existing file_names in expressions.
expand_file_names () { eval $LD "$@"; }

# Get the absolute path to a given file path.
# We should use the following, providing the realpath command with option -s is
# available under BSD...
#    $REALPATH `expand "$1"`
# This one is called at least by absolute_command_search_path after having
# set IFS to ":".
absolute_file_path () { unset IFS; expand "$1"; }

# Local function.
check_bool () {
    IDENT="$1"
    VALUE="$2"
    
    case "${VALUE:?}" in
    ( true | false ) :;;
    ( * )
        error "$IDENT = \"$VALUE\" is not a boolean value"
        exit ${INVALID_ARGUMENT_ERROR:?}
        ;;
    esac
}

# Local function.
check_not_here () {
    IDENT="$1"
    VALUE="$2"
    
    HERE=`absolute_file_path "."` && \
    VALUE=`absolute_file_path "$VALUE"`
    IER=$?
    if test $IER -ne 0; then exit $IER; fi
    if test "$VALUE" = "$HERE"; then
        error "$IDENT cannot be the current directory"
        exit ${INVALID_VALUE_ERROR:?}
    fi
}

# Caml bindings of the following form:
# <let> <space> <ident> <space> <equal> <space> <value> <semisemi>
get_caml_binding_from_file () {
    FILE="$1"
    IDENT="$2"
    
    $CAT "${FILE:?}" | \
    $EGREP "^let $IDENT = .*;;\$" | \
    $AWK -F" = " '{print $2}' | \
    strip_last_suffix ";;"
}

replace_caml_binding_in_file () {
    FILE="$1"
    IDENT="$2"
    VALUE="$3"
    
    $CAT "${FILE:?}" | \
    $SED -e "/^let $IDENT = .*;;\$/s@ = .*;;@ = $VALUE;;@" 1> "$FILE.tmp"
    IER=$?
    if test $IER -ne 0; then exit $IER; fi
    $MV "$FILE.tmp" "$FILE"
}

#
# Files.
#

# Symbolic links.

# Local function.
get_link_source () {
    TARGET="$1"
    
    $LLD "${TARGET:?}" | \
    $AWK -F"->" '{print $2}' | \
    $STRIP_SPACE
}

# This one is called at least by locate_command after having set IFS to ":".
# Local function.
dereference_link () {
    unset IFS
    FILE="$1"
    
    while test -h "${FILE:?}"; do
        NEW_FILE=`get_link_source "$FILE"`
        IER=$?
        if test $IER -ne 0; then exit $IER; fi
        case "${NEW_FILE:?}" in
        ( /* ) FILE="$NEW_FILE";;
        ( * )
            DIR=`$DIRNAME "$FILE"`
            FILE="$DIR/$NEW_FILE"
            ;;
        esac
    done
    $ECHO "$FILE"
}

create_link () {
    SUB_DIR="$1"
    SOURCE="$2"
    TARGET="$3"
    
    (\cd "${SUB_DIR:?}" 1> /dev/null && \
     $RM "$TARGET" && \
     $LN "${SOURCE:?}" "${TARGET:?}")
}

# Unused.
#safe_create_link () {
#    SUB_DIR="$1"
#    SOURCE="$2"
#    TARGET="$3"
#    
#    (\cd "${SUB_DIR:?}" 1> /dev/null
#    if ! test -e "${TARGET:?}"; then
#        $LN "${SOURCE:?}" "$TARGET"
#    elif ! test -h "$TARGET"; then
#        MY_FILE="file \"$SUB_DIR/$TARGET\""
#        error "$MY_FILE is already existing but it is not a symbolic link."
#        exit ${WRONG_FILE_TYPE_ERROR:?}
#    else
#        LNK=`get_link_source "$TARGET"`
#        IER=$?
#        if test $IER -ne 0; then exit $IER; fi
#        if test "$LNK" != "$SOURCE"; then
#            MY_FILE="file \"$SUB_DIR/$TARGET\""
#            error "$MY_FILE is a symbolic link towards \"$LNK\"" \
#                  "instead of \"$SOURCE\"."
#            exit ${WRONG_LINK_ERROR:?}
#        fi
#    fi)
#}

# Replace file only if it is different.
# Local function.
replace_if_new () {
    FILE1="$1"
    FILE2="$2"
    
    TO_DO="$MV \"${FILE1:?}\" \"${FILE2:?}\""
    if test -f "${FILE2:?}"; then
        $CMP "${FILE1:?}" "${FILE2:?}"
        IER=$?
        case $IER in
        ( 0 ) TO_DO="$RM \"$FILE1\"";;
        ( 1 ) :;;
        ( * ) exit $IER;;
        esac
    fi
    eval "$TO_DO"
}

# Local function.
replace_if_new_ro () {
    FILE1="$1"
    FILE2="$2"
    
    replace_if_new "$FILE1" "$FILE2"
    $CHMOD_READ_ONLY "${FILE2:?}"
}

# Remove the directory and its parents only if they are empty.
# And dont bother about error messages nor nonzero exit codes.
# Local function.
remove_empty_dir_and_parents () {
    $RMDIR_AND_PARENTS "$1" 2> /dev/null || true
}

remove_pph_private_install_dir () {
    remove_empty_dir_and_parents "${PPH_DOC_DIR:?}" && \
    remove_empty_dir_and_parents "${PPH_MAN_DIR:?}" && \
    remove_empty_dir_and_parents "${PPH_SHR_DIR:?}" && \
    remove_empty_dir_and_parents "${PPH_LIB_DIR:?}" && \
    remove_empty_dir_and_parents "${PPH_INC_DIR:?}" && \
    remove_empty_dir_and_parents "${PPH_BIN_DIR:?}" && \
    remove_empty_dir_and_parents "${PPH_PRIVATE_ROOT_DIR:?}"
}

# Unused.
#strip_cvsid () {
#    NEWMSG="$1"
#    FILEIN="$2"
#    FILEOUT="$3"
#    
#    ID="\\\$Id:"
#    EXP="Exp \\\$"
#    case "${FILEIN:?}" in
#    ( - ) TO_DO="$SED -e 's/$ID .*$EXP/$NEWMSG/'";;
#    ( * ) TO_DO="$SED -e 's/$ID .*$EXP/$NEWMSG/' \"$FILEIN\"";;
#    esac
#    case "${FILEOUT:?}" in
#    ( - ) eval "$TO_DO";;
#    ( * )
#        TMPFILEOUT="$FILEOUT".tmp_strip
#        eval "$TO_DO" 1> "$TMPFILEOUT"
#        replace_if_new "$TMPFILEOUT" "$FILEOUT"
#        ;;
#    esac
#}

# Search path for commands.
get_command_search_path () {
    COMMAND_SEARCH_PATH="$PATH"
    
    for ROOT_DIR in ${PATH_ROOT_DIRS_INTRINSIC_DEFAULT:?}; do
        case "${ROOT_DIR:?}" in
        ( */ ) :;;
        ( * ) ROOT_DIR="$ROOT_DIR/";;
        esac
        for SUB_DIR in ${PATH_SUB_DIRS_INTRINSIC_DEFAULT:?}; do
            DIR=`expand "$ROOT_DIR$SUB_DIR"`
            IER=$?
            if test $IER -ne 0; then exit $IER; fi
            if ! is_in_colon_list "$COMMAND_SEARCH_PATH" "$DIR"; then
                if test -d "${DIR:?}"; then
                    if test -n "$COMMAND_SEARCH_PATH"; then
                        COMMAND_SEARCH_PATH="$COMMAND_SEARCH_PATH:$DIR"
                    else
                        COMMAND_SEARCH_PATH="$DIR"
                    fi
                fi
            fi
        done
    done
    $ECHO "$COMMAND_SEARCH_PATH"
}

# List of absolute paths.
absolute_command_search_path () {
    COMMAND_SEARCH_PATH="$1"
    
    RES=""
    # Set the field separator to colon (':').
    # Be careful as this may perturbate other pieces of code (space is no
    # more a separator), consider issuing "unset IFS" there.
    IFS=":"
    for DIR in $COMMAND_SEARCH_PATH; do
        DIR=`absolute_file_path "$DIR"`
        IER=$?
        if test $IER -ne 0; then exit $IER; fi
        if test -n "$RES"; then
            RES="$RES:$DIR"
        else
            RES="$DIR"
        fi
    done
    unset IFS
    $ECHO "$RES"
}

# Local function.
get_version () {
    COMMAND="$1"
    
    case "${COMMAND:?}" in
    ( *cpp* | *gcc* | *g++* | *gcj** | *icc* | *icpc* ) FLAGS="-dumpversion";;
    ( *gfortran* ) FLAGS="-dumpversion | $HEAD 1 | $AWK '{print \$NF}'";;
    ( *ifort* ) FLAGS="--version | $HEAD 1 | $AWK '{print \$3}'";;
    ( *mocac* ) FLAGS="-v 2>&1 | $HEAD 1 | $AWK '{print \$NF}'";;
    ( *camlp5* ) FLAGS="-v 2>&1 | $HEAD 1 | $AWK '{print \$3}'";;
    ( *camlp4* | *ocamlc* | *ocamlopt* ) FLAGS="-vnum";;
    ( * )
        error "don't know how to get the version number of \"$COMMAND\""
        exit ${NOT_YET_IMPLEMENTED_ERROR:?}
        ;;
    esac
    eval "$COMMAND $FLAGS"
}

# Local function.
get_version_field () {
    NUMBER="$1"
    LINE="$2"
    
    FIELD=`$ECHO "$LINE" | $CUT -d . -f ${NUMBER:?} | \
           $AWK -F- '{print $1}' | $AWK -F_ '{print $1}'`
    if test -z "$FIELD"; then
        FIELD="0"
    fi
    $ECHO "$FIELD"
}

# Local functions.
get_major () { get_version_field 1 "$1"; }
get_minor () { get_version_field 2 "$1"; }
get_patch () { get_version_field 3 "$1"; }

# Local function.
check_order () {
    VALUE="$1"
    SPEC_OPERATOR="$2"
    SPEC_VALUE="$3"
    
    case "${SPEC_OPERATOR:?}" in
    ( "<"* )
        case "${SPEC_OPERATOR:?}" in
        ( "<=" ) OPPOSITE_SPEC_OPERATOR=">";;
        ( "<" ) OPPOSITE_SPEC_OPERATOR=">=";;
        ( "<>" ) OPPOSITE_SPEC_OPERATOR="=";;
        ( * )
            error "unknown operator \"$SPEC_OPERATOR\""
            exit ${INVALID_ARGUMENT_ERROR:?}
            ;;
        esac
        DIAG=`check_order "$VALUE" "$OPPOSITE_SPEC_OPERATOR" "$SPEC_VALUE"`
        IER=$?
        if test $IER -ne 0; then exit $IER; fi
        case "${DIAG:?}" in
        ( OK ) $ECHO "KO";;
        ( KO ) $ECHO "OK";;
        ( * )
            internal_error "should be OK or KO, not \"$DIAG\""
            exit ${INTERNAL_ERROR:?}
            ;;
        esac
        ;;
    ( "=" )
        if test "$VALUE" = "$SPEC_VALUE"; then
            $ECHO "OK"
        else
            $ECHO "KO"
        fi
        ;;
    ( ">"* )
        MAJOR=`get_major "$VALUE"` && \
        MINOR=`get_minor "$VALUE"` && \
        PATCH=`get_patch "$VALUE"` && \
        SPEC_MAJOR=`get_major "$SPEC_VALUE"` && \
        SPEC_MINOR=`get_minor "$SPEC_VALUE"` && \
        SPEC_PATCH=`get_patch "$SPEC_VALUE"`
        IER=$?
        if test $IER -ne 0; then exit $IER; fi
        if test "$MAJOR" -lt "$SPEC_MAJOR"; then
            $ECHO "KO"
        elif test "$MAJOR" -gt "$SPEC_MAJOR"; then
            $ECHO "OK"
        else
            if test "$MINOR" -lt "$SPEC_MINOR"; then
                $ECHO "KO"
            elif test "$MINOR" -gt "$SPEC_MINOR"; then
                $ECHO "OK"
            else
                if test "$PATCH" -lt "$SPEC_PATCH"; then
                    $ECHO "KO"
                elif test "$PATCH" -gt "$SPEC_PATCH"; then
                    $ECHO "OK"
                else
                    case "${SPEC_OPERATOR:?}" in
                    ( ">" ) $ECHO "KO";;
                    ( ">=" ) $ECHO "OK";;
                    ( * )
                        error "unknown operator \"$SPEC_OPERATOR\""
                        exit ${INVALID_ARGUMENT_ERROR:?}
                        ;;
                    esac
                fi
            fi
        fi
        ;;
    ( * )
        error "unknown operator \"$SPEC_OPERATOR\""
        exit ${INVALID_ARGUMENT_ERROR:?}
        ;;
    esac
}

# This one is called at least by locate_command after having set IFS to ":".
# Local function.
check_version_spec () {
    unset IFS
    COMMAND="$1"
    SPEC_OPERATOR="$2"
    SPEC_VALUE="$3"
    
    VALUE=`get_version "$COMMAND"`
    IER=$?
    if test $IER -ne 0; then exit $IER; fi
    check_order "$VALUE" "$SPEC_OPERATOR" "$SPEC_VALUE"
}

# Locate the first executable occurrence of a command (second argument) in the
# list of colon (':') separated directory paths (first argument).
# Returns "Not_found" if the command is not found, or not executable.
#
# If you want to modify this, be aware that:
#  * ``which'' and ``whereis'' may fail, do not use the same options
#    everywhere, and may even not be available,
#  * ``type'' do not use the same options everywhere, and may even not be
#    available (e.g. its equivalent in Bourne Shell under OpenBSD is the
#    builtin ``whence'').
# Local function.
locate_command () {
    COMMAND_SEARCH_PATH="$1"
    COMMAND_SPEC="$2"
    
    COMMAND_SPEC_NB_FIELDS=`$ECHO "$COMMAND_SPEC" | $AWK -F"@@" '{print NF}'`
    case "${COMMAND_SPEC_NB_FIELDS:?}" in
    ( 1 ) COMMAND="$COMMAND_SPEC";;
    ( 3 )
        COMMAND=`$ECHO "$COMMAND_SPEC" | $AWK -F"@@" '{print $1}'`
        SPEC_OPE=`$ECHO "$COMMAND_SPEC" | $AWK -F"@@" '{print $2}'`
        SPEC_VAL=`$ECHO "$COMMAND_SPEC" | $AWK -F"@@" '{print $3}'`
        ;;
    ( * )
        error "command spec must be of the form \"<command>[@@<ope>@@<val>]\""
        exit ${INVALID_ARGUMENT_ERROR:?}
        ;;
    esac
    
    CMD="Not_found"
    # Set the field separator to colon (':').
    # Be careful as this may perturbate other pieces of code (space is no
    # more a separator), consider issuing "unset IFS" there.
    IFS=":"
    for DIR in $COMMAND_SEARCH_PATH; do
        RES=`dereference_link "$DIR/$COMMAND"`
        IER=$?
        if test $IER -ne 0; then exit $IER; fi
        if test -x "${RES:?}"; then
            if test "${COMMAND_SPEC_NB_FIELDS:?}" -eq 1; then
                CMD="$RES"
                break
            else
                DIAG=`check_version_spec "$RES" "$SPEC_OPE" "$SPEC_VAL"`
                IER=$?
                if test $IER -ne 0; then exit $IER; fi
                case "${DIAG:?}" in
                ( OK ) CMD="$RES"; break;;
                ( KO ) continue;;
                esac
            fi
        fi
    done
    unset IFS
    
    $ECHO "$CMD"
}

# Find the first available executable occurrence of a command in the list
# given as the third argument. The possibly empty second argument allows to
# specify a list of filename extensions that are also searched for (this is
# useful when an optimized version is also available).
# The first argument is the command search path, i.e. a list of colon (':')
# separated directory paths.
# Local function.
find_first_command () {
    COMMAND_SEARCH_PATH="$1"
    SUFFIXES="$2"
    LIST="$3"
    
    CMD="Not_found"
    for ITEM in $LIST; do
        RES=`locate_command "$COMMAND_SEARCH_PATH" "$ITEM"`
        IER=$?
        if test $IER -ne 0; then exit $IER; fi
        case "${RES:?}" in
        ( Not_found ) :;;
        ( /* )
            if test -n "$SUFFIXES"; then
                for SUFFIX in $SUFFIXES; do
                    if test -x "$RES$SUFFIX"; then
                        RES="$RES$SUFFIX"
                        break
                    fi
                done
            fi
            CMD="$RES"
            break
            ;;
        ( * )
            internal_error "\"locate_command\" cannot return \"$RES\""
            exit ${INTERNAL_ERROR:?}
            ;;
        esac
    done
    
    $ECHO "$CMD"
}

get_command () {
    STATUS="$1"
    INTRINSIC_DEFAULT_LIST="$2"
    CURRENT="$3"
    FROM_COMMAND_LINE="$4"
    COMMAND_SEARCH_PATH="$5"
    SUFFIXES="$6"
    
    case "${STATUS:?}" in
    ( None ) :;;
    ( Mandatory | Optional )
        LIST="$INTRINSIC_DEFAULT_LIST"
        if test "$CURRENT" != "False"; then
            LIST=`$BASENAME $CURRENT`
        fi
        if test -n "$FROM_COMMAND_LINE"; then
            LIST="$FROM_COMMAND_LINE"
        fi
        CURRENT=`\
            find_first_command "$COMMAND_SEARCH_PATH" "$SUFFIXES" "$LIST"`
        IER=$?
        if test $IER -ne 0; then exit $IER; fi
        case "${STATUS:?}" in
        ( Mandatory )
            if test "$CURRENT" = "Not_found"; then
                if test `$ECHO "$LIST" | $WC_WORD` -eq 1; then
                    MSG="The command \"$LIST\" has not been found
in the following path list:"
                    REM_MSG="Install it..."
                else
                    MSG="None of the commands \"$LIST\" has been found
in the following path list:"
                    REM_MSG="Install one of them..."
                fi
                IFS=":"
                for DIR in $COMMAND_SEARCH_PATH; do
                    MSG="$MSG
$DIR"
                done
                unset IFS
                error "$MSG"
                remedy "$REM_MSG"
                exit ${COMMAND_NOT_FOUND_ERROR:?}
            fi
            ;;
        ( Optional ) :;;
        ( * )
            internal_error "go away, you dont exist"
            exit ${PPH_INTERNAL_ERROR:?}
            ;;
        esac
        ;;
    ( * )
        error "unknown command status \"$STATUS\""
        exit ${INVALID_ARGUMENT_ERROR:?}
        ;;
    esac
    echo "$CURRENT"
}

# Record a variable assignment in a given configuration file.
# We use here the makefile syntax.
# Local function.
record_variable_assignment_for_makefile () {
    CONFIG_FILE="$1"
    COMMENT="$2"
    IDENT="$3"
    VALUE=`expand "$4"`
    
    $CAT << eof >> "${CONFIG_FILE:?}"

# $COMMENT.
$IDENT = $VALUE
eof
}

# For shell, we need double quotes to allow for white spaces in values.
# Local function.
record_variable_assignment_for_shell () {
    CONFIG_FILE="$1"
    COMMENT="$2"
    IDENT="$3"
    VALUE=`expand "$4"`
    
    $CAT << eof >> "${CONFIG_FILE:?}"

# $COMMENT.
$IDENT="$VALUE"
eof
}

# Local function.
record_variable_assignment_for_htmlc_environment () {
    CONFIG_FILE="$1"
    COMMENT="$2"
    IDENT="$3"
    VALUE=`expand "$4"`
    
    $CAT << eof >> "${CONFIG_FILE:?}"

(* $COMMENT. *)
let $IDENT = "$VALUE";;
eof
}

# Initiate the filling of a configuration file
init_config_file () {
    CONFIG_FILE_NO_EXTENSION="$1"
    
    FILE_BASENAME=`$BASENAME "$CONFIG_FILE_NO_EXTENSION"`
    FILE_HTMLC_ENV="$CONFIG_FILE_NO_EXTENSION.env"
    FILE_MAKEFILE="$CONFIG_FILE_NO_EXTENSION.pph"
    FILE_SHELL="$CONFIG_FILE_NO_EXTENSION.profile"
    TMP_FILE_HTMLC_ENV="$FILE_HTMLC_ENV.tmp_fill"
    TMP_FILE_MAKEFILE="$FILE_MAKEFILE.tmp_fill"
    TMP_FILE_SHELL="$FILE_SHELL.tmp_fill"
    MSG_HTMLC_ENV=`gen_do_warning_msg_for_echo` && \
    MSG_MAKEFILE=`gen_dont_warning_msg_for_echo "$FILE_BASENAME.env"` && \
    MSG_SHELL=`gen_dont_warning_msg_for_echo "$FILE_BASENAME.env"`
    IER=$?
    if test $IER -ne 0; then exit $IER; fi
    $ECHO "$MSG_HTMLC_ENV" 1> "$TMP_FILE_HTMLC_ENV"
    $ECHO "$MSG_MAKEFILE" 1> "$TMP_FILE_MAKEFILE"
    $ECHO "$MSG_SHELL" 1> "$TMP_FILE_SHELL"
}

# Add a list of comment/variable/value triples to an already initiated
# configuration file.
add_to_config_file () {
    CONFIG_FILE_NO_EXTENSION="$1"
    
    FILE_HTMLC_ENV="$CONFIG_FILE_NO_EXTENSION.env"
    FILE_MAKEFILE="$CONFIG_FILE_NO_EXTENSION.pph"
    FILE_SHELL="$CONFIG_FILE_NO_EXTENSION.profile"
    TMP_FILE_HTMLC_ENV="$FILE_HTMLC_ENV.tmp_fill"
    TMP_FILE_MAKEFILE="$FILE_MAKEFILE.tmp_fill"
    TMP_FILE_SHELL="$FILE_SHELL.tmp_fill"
    shift
    while test $# -ge 3; do
        VARIABLE_NAME="$1"
        COMMENT="Binding for $VARIABLE_NAME"
        IDENT="$2"
        VALUE="$3"
        shift 3
        if test "${VALUE:?}" != "False"; then
            record_variable_assignment_for_htmlc_environment \
                "$TMP_FILE_HTMLC_ENV" "$COMMENT" "$IDENT" "$VALUE"
            record_variable_assignment_for_makefile \
                "$TMP_FILE_MAKEFILE" "$COMMENT" "$IDENT" "$VALUE"
            record_variable_assignment_for_shell \
                "$TMP_FILE_SHELL" "$COMMENT" "$IDENT" "$VALUE"
            eval "$IDENT=\"$VALUE\""
        fi
    done
    if test $# -ge 1; then
        internal_error "missing ident and/or value for variable $1"
        exit ${INTERNAL_ERROR:?}
    fi
}

# Process configuration options.
set_configuration_binding () {
    OPTION_NAME="$1"
    IDENT="$2"
    KIND="$3"
    CONFIG_FILE_NO_EXTENSION="$4"
    INTRINSIC_DEFAULT="$5"
    FROM_CONFIG_FILE="$6"
    FROM_COMMAND_LINE="$7"
    
    CONFIG_FILE_BNAME=`$BASENAME "$CONFIG_FILE_NO_EXTENSION"`
    
    report_message "Setting $OPTION_NAME binding in $CONFIG_FILE_BNAME"
    
    # Intrinsic default value.
    CANDIDATE=`expand "$INTRINSIC_DEFAULT"`
    IER=$?
    if test $IER -ne 0; then exit $IER; fi
    
    # Value from the project configuration file,
    # or computed from the current environment.
    if test -n "$FROM_CONFIG_FILE"; then
        CANDIDATE=`expand "$FROM_CONFIG_FILE"`
        IER=$?
        if test $IER -ne 0; then exit $IER; fi
    else
        case "${KIND:?}" in
        ( *versioning_system* )
            if test -f "./CVS/Entries"; then
                CANDIDATE="CVS"
            elif test -f "./.svn/entries"; then
                CANDIDATE="Subversion"
            else
                CANDIDATE="None"
            fi
            ;;
        ( *tarballs_install_root_dir* | *tarballs_mimic_install_root_dir* )
            case "${KIND:?}" in
            ( *tarballs_install_root_dir* )
                CANDIDATE="${INSTALL_ROOT_DIR:?}";;
            ( *tarballs_mimic_install_root_dir* )
                CANDIDATE="${MIMIC_INSTALL_ROOT_DIR:?}";;
            esac
            if ! ${PRIVATE_TOOL:?}; then
                PPSD="${PROJECT_INSTALLATION_NAME:?}/private"
                MAIN_LANGUAGE="${PROJECT_MAIN_LANGUAGE:?}"
                case "${MAIN_LANGUAGE:?}" in
                ( c | c++ | fortran | java | ocaml )
                    PISD="lib/$MAIN_LANGUAGE/$PPSD"
                    ;;
                ( text ) PISD="share/$PPSD";;
                esac
                CANDIDATE="${CANDIDATE:?}/${PISD:?}"
            fi
            ;;
        esac
    fi
    
    # Value from the command line.
    if test -n "$FROM_COMMAND_LINE"; then
        CANDIDATE=`expand "$FROM_COMMAND_LINE"`
        IER=$?
        if test $IER -ne 0; then exit $IER; fi
    fi
    
    # Perform sanity checks and normalizations.
    if test -z "$CANDIDATE"; then
        error "empty $OPTION_NAME"
        exit ${EMPTY_VALUE_ERROR:?}
    fi
    case "${KIND:?}" in
    ( *bool* )
        check_bool "$IDENT" "$CANDIDATE"
        IER=$?
        if test $IER -ne 0; then exit $IER; fi
        ;;
    esac
    case "${KIND:?}" in
    ( *absolute_file_path* )
        CANDIDATE=`absolute_file_path "$CANDIDATE"`
        IER=$?
        if test $IER -ne 0; then exit $IER; fi
        ;;
    esac
    case "${KIND:?}" in
    ( *not_here* )
        check_not_here "$IDENT" "$CANDIDATE"
        IER=$?
        if test $IER -ne 0; then exit $IER; fi
        ;;
    esac
    case "${KIND:?}" in
    ( *versioning_system* )
        case "${CANDIDATE:?}" in
        ( CVS )
            VERSIONING_ADD="$CVS_ADD"
            VERSIONING_COMMIT="$CVS_COMMIT"
            VERSIONING_REMOVE="$CVS_REMOVE"
            ;;
        ( Subversion )
            VERSIONING_ADD="$SVN_ADD"
            VERSIONING_COMMIT="$SVN_COMMIT"
            VERSIONING_REMOVE="$SVN_DELETE"
            ;;
        ( * )
            VERSIONING_ADD="$ECHO add"
            VERSIONING_COMMIT="$ECHO commit"
            VERSIONING_REMOVE="$ECHO remove"
            ;;
        esac
        ;;
    esac
    
    # Store the current setting.
    add_to_config_file \
        "$CONFIG_FILE_NO_EXTENSION" \
        "$OPTION_NAME" "$IDENT" "$CANDIDATE"
    IER=$?
    if test $IER -ne 0; then exit $IER; fi
    case "${KIND:?}" in
    ( *versioning_system* )
        add_to_config_file \
            "$CONFIG_FILE_NO_EXTENSION" \
            "Versioning add" "    VERSIONING_ADD" "$VERSIONING_ADD" \
            "Versioning commit" "    VERSIONING_COMMIT" "$VERSIONING_COMMIT" \
            "Versioning remove" "    VERSIONING_REMOVE" "$VERSIONING_REMOVE"
        IER=$?
        if test $IER -ne 0; then exit $IER; fi
        ;;
    esac
}

# Register an already filled configuration file.
# An existing file is only regenerated when the contents would be different.
register_config_file () {
    CONFIG_FILE_NO_EXTENSION="$1"
    
    FILE_HTMLC_ENV="$CONFIG_FILE_NO_EXTENSION.env"
    FILE_MAKEFILE="$CONFIG_FILE_NO_EXTENSION.pph"
    FILE_SHELL="$CONFIG_FILE_NO_EXTENSION.profile"
    TMP_FILE_HTMLC_ENV="$FILE_HTMLC_ENV.tmp_fill"
    TMP_FILE_MAKEFILE="$FILE_MAKEFILE.tmp_fill"
    TMP_FILE_SHELL="$FILE_SHELL.tmp_fill"
    replace_if_new "$TMP_FILE_HTMLC_ENV" "$FILE_HTMLC_ENV"
    replace_if_new_ro "$TMP_FILE_MAKEFILE" "$FILE_MAKEFILE"
    replace_if_new_ro "$TMP_FILE_SHELL" "$FILE_SHELL"
}

#
# Documentation.
#

# Local function.
gen_doc () {
    LONG_OPTION_KEY="$1"
    SHORT_OPTION_KEY="$2"
    ARGUMENT="$3"
    TEXT="$4"
    
    $ECHO "$LONG_OPTION_KEY$ARGUMENT, $SHORT_OPTION_KEY$ARGUMENT
		$TEXT."
}

# Local function.
gen_doc_with_argument () {
    LONG_OPTION_KEY="$1"
    SHORT_OPTION_KEY="$2"
    ARGUMENT_NAME="$3"
    ARGUMENT=" <$ARGUMENT_NAME>"
    TEXT="$4"
    
    gen_doc "$LONG_OPTION_KEY" "$SHORT_OPTION_KEY" "$ARGUMENT" "$TEXT"
}

doc () {
    LONG_OPTION_KEY="$1"
    SHORT_OPTION_KEY="$2"
    ARGUMENT_NAME="$3"
    ARGUMENT_VALUE="<${ARGUMENT_NAME:?}>"
    OPTION_NAME="$4"
    
    gen_doc_with_argument \
        "${LONG_OPTION_KEY:?}" "${SHORT_OPTION_KEY:?}" "$ARGUMENT_NAME" \
        "Set the ${OPTION_NAME} to $ARGUMENT_VALUE"
}

doc_with_default () {
    LONG_OPTION_KEY="$1"
    SHORT_OPTION_KEY="$2"
    ARGUMENT_NAME="$3"
    ARGUMENT_VALUE="<${ARGUMENT_NAME:?}>"
    OPTION_NAME="$4"
    DEFAULT_VALUE="$5"
    
    gen_doc_with_argument \
        "${LONG_OPTION_KEY:?}" "${SHORT_OPTION_KEY:?}" "$ARGUMENT_NAME" \
        "Set the ${OPTION_NAME:?} to $ARGUMENT_VALUE.
		Default is \"${DEFAULT_VALUE:?}\""
}

# Unused.
#doc_toggle () {
#    LONG_OPTION_KEY="$1"
#    SHORT_OPTION_KEY="$2"
#    OPTION_NAME="$3"
#    NEW_VALUE="$4"
#    DEFAULT_VALUE="$5"
#    
#    gen_doc \
#        "${LONG_OPTION_KEY:?}" "${SHORT_OPTION_KEY:?}" "" \
#        "Set the ${OPTION_NAME:?} to \"${NEW_VALUE:?}\".
#		Default is \"${DEFAULT_VALUE:?}\""
#}

# Usage header.
print_usage_header () {
    HEADER_BODY="
Usage: ./$SCRIPT [<option>...]

OPTIONS"
    $ECHO "$HEADER_BODY"
}

# Usage section.
print_usage_section () {
    TITLE="$1"
    BODY="$2"
    
    $ECHO "
$TITLE:
"
    $ECHO "$BODY"
}

# Service section.
print_service_section () {
    SERVICE_TITLE="Service"
    SERVICE_BODY="\
--trace, -x	Trigger the fine debugging mode on. Default is "${TRACE:?}".
--debug, -d	Trigger the debug mode on. Default is "${DEBUG:?}".
--report, -r	Trigger the reporting mode on. Default is "${REPORT:?}".
--summary, -s	Trigger the summary mode on. Default is "${SUMMARY:?}".
--see-you, -n	Trigger the see-you mode on. Default is "${SEE_YOU:?}".

--help, -h	Display this help and exit.
"
    print_usage_section "$SERVICE_TITLE" "$SERVICE_BODY"
}

#
# Path files.
#

# Create path files in the root directory of the project.
create_path_file () {
    PROJECT_SUB_DIR="$1"
    PROJECT_ROOT_DIR_RELATIVE_PATH="$2"
    PATH_FILE="$3"
    PATH_FILE="${PROJECT_ROOT_DIR:?}/${PROJECT_SUB_DIR:?}/${PATH_FILE:?}"
    
    $RM "$PATH_FILE"
    run_command "$CAT << eof 1> \"${PATH_FILE:?}\"
PROJECT_ROOT_DIR=$PROJECT_ROOT_DIR
PROJECT_SUB_DIR=$PROJECT_SUB_DIR
PROJECT_ROOT_DIR_RELATIVE_PATH=$PROJECT_ROOT_DIR_RELATIVE_PATH
eof"
    $CHMOD_READ_ONLY "$PATH_FILE"
}

# Create path files in the source directory of Pph pervasive tools.
create_path_file_in_pph_pervasive_tool () {
    TOOL="$1"
    
    TOOL_RELATIVE_PATH="pph/$TOOL"
    TOOL_PARENT_DIR="../.."
    if test -d "$TOOL_RELATIVE_PATH"; then
        create_path_file \
            "$TOOL_RELATIVE_PATH" \
            "$TOOL_PARENT_DIR" \
            "${PROJECT_PATH_FILE:?}"
    fi
}

create_path_file_in_pph_pervasive_tools () {
    TOOLS="$1"
    
    LIST=`$ECHO "$TOOLS" | $STRIP_SPACE`
    if test -z "$LIST"; then return 0; fi
    for TOOL in $TOOLS; do
        create_path_file_in_pph_pervasive_tool "$TOOL"
        IER=$?
        if test $IER -ne 0; then exit $IER; fi
    done
}

# Remove path files in the source directory of Pph pervasive tools.
remove_path_file_in_pph_pervasive_tool () {
    TOOL="$1"
    
    TOOL_RELATIVE_PATH="pph/$TOOL"
    if test -d "$TOOL_RELATIVE_PATH"; then
        report_message "Removing project path in \"$TOOL\""
        $RM "$TOOL_RELATIVE_PATH/${PROJECT_PATH_FILE:?}"
    fi
}

remove_path_file_in_pph_pervasive_tools () {
    TOOLS="$1"
    
    LIST=`$ECHO "$TOOLS" | $STRIP_SPACE`
    if test -z "$LIST"; then return 0; fi
    for TOOL in $TOOLS; do
        remove_path_file_in_pph_pervasive_tool "$TOOL"
        IER=$?
        if test $IER -ne 0; then exit $IER; fi
    done
}

#
# Preambles.
#

# Local function.
create_preamble () {
    PREAMBLE_TYPE="$1"
    PREAMBLE_FILE="$2"
    PREAMBLE_FILE="${PROJECT_ROOT_DIR:?}/${PREAMBLE_FILE:?}"
    PREAMBLE_INIT="$3"
    
    $RM "$PREAMBLE_FILE"
    run_command "$CAT << eof 1> \"${PREAMBLE_FILE:?}\"
include ./$PREAMBLE_INIT
include \\\$(PROJECT_ROOT_DIR)/pph/$PREAMBLE_TYPE/makefiles/Pph_dirs.pph
eof"
    $CHMOD_READ_ONLY "$PREAMBLE_FILE"
}

create_user_preamble () { create_preamble "user" "$1" "$2"; }
create_devel_preamble () { create_preamble "devel" "$1" "Makefile_user"; }

# Pph pervasive tools.
# Installation immediately follows compilation,
# hence we can use a volatile location for compilation as /tmp.
# Installation goes into $PPH_SYSTEM_ROOT_DIR.
# Local function.
install_pph_pervasive_tool () {
    FORCE="$1"
    TOOL="$2"
    
    SRC_DIR="${PPH_SYSTEM_ROOT_DIR:?}/$TOOL"
    TARBALL="${PPH_SYSTEM_ROOT_DIR:?}/tarballs/$TOOL.tgz"
    
    if ! test -d "$SRC_DIR" && ! test -f "$TARBALL"; then
        error "no sources to build and install \"$TOOL\""
        exit ${INVALID_ARGUMENT_ERROR:?}
    fi
    
    TO_DO=true
    if test -x "${PPH_BIN_DIR:?}/$TOOL"; then
        TO_DO=false
    fi
    if ${FORCE:?}; then
        TO_DO=true
    fi
    
    if $TO_DO; then
        report_message "Installing Pph pervasive tool \"$TOOL\""
        if test -d "$SRC_DIR"; then
            (\cd "${SRC_DIR:?}" && \
             $SH ./build.sh && \
             $SH ./install.sh "${PPH_BIN_DIR:?}" && \
             $SH ./clean.sh)
        elif test -f "$TARBALL"; then
            TMP_DIR=`$MK_TEMP_DIR`
            $RM "$TMP_DIR" && \
            $MKDIR "$TMP_DIR" && \
            (\cd "${TMP_DIR:?}" 1> /dev/null && \
             $TARXZ "$TARBALL" && \
             \cd "${TOOL:?}"-* 1> /dev/null && \
             ./configure --install-root-dir "${PPH_PRIVATE_ROOT_DIR:?}" && \
             $MAKE_COMMAND && \
             $MAKE_COMMAND install) && \
            $RM "$TMP_DIR"
        fi
    else
        report_message "Pph pervasive tool \"$TOOL\" is already installed"
    fi
}

install_pph_pervasive_tools () {
    FORCE="$1"
    TOOLS="$2"
    
    LIST=`$ECHO "$TOOLS" | $STRIP_SPACE`
    if test -z "$LIST"; then return 0; fi
    for TOOL in $TOOLS; do
        install_pph_pervasive_tool "$FORCE" "$TOOL"
        IER=$?
        if test $IER -ne 0; then exit $IER; fi
    done
}

# Local function.
uninstall_pph_pervasive_tool () {
    TOOL="$1"
    
    SRC_DIR="${PPH_SYSTEM_ROOT_DIR:?}/$TOOL"
    TARBALL="${PPH_SYSTEM_ROOT_DIR:?}/tarballs/$TOOL.tgz"
    
    if ! test -d "$SRC_DIR" && ! test -f "$TARBALL"; then
        error "no sources to build and install \"$TOOL\""
        exit ${INVALID_ARGUMENT_ERROR:?}
    fi
    
    TO_DO=true
    if ! test -x "${PPH_BIN_DIR:?}/$TOOL"; then
        TO_DO=false
    fi
    
    if $TO_DO; then
        report_message "Uninstalling Pph pervasive tool \"$TOOL\""
        if test -d "$SRC_DIR"; then
            (\cd "${SRC_DIR:?}" 1> /dev/null && \
             $SH ./uninstall.sh "${PPH_BIN_DIR:?}") 1> /dev/null 2>&1
        elif test -f "$TARBALL"; then
            TMP_DIR=`$MK_TEMP_DIR`
            $RM "$TMP_DIR" && \
            $MKDIR "$TMP_DIR" && \
            (\cd "${TMP_DIR:?}" 1> /dev/null && \
             $TARXZ "$TARBALL" && \
             \cd "${TOOL:?}"-* 1> /dev/null && \
             ./configure --install-root-dir "${PPH_PRIVATE_ROOT_DIR:?}" && \
             $MAKE_COMMAND uninstall) 1> /dev/null 2>&1 && \
            $RM "$TMP_DIR" && \
            remove_pph_private_install_dir
        fi
    else
        report_message "Pph pervasive tool \"$TOOL\" is not installed"
    fi
}

uninstall_pph_pervasive_tools () {
    TOOLS="$1"
    
    LIST=`$ECHO "$TOOLS" | $STRIP_SPACE`
    if test -z "$LIST"; then return 0; fi
    for TOOL in $TOOLS; do
        uninstall_pph_pervasive_tool "$TOOL"
        IER=$?
        if test $IER -ne 0; then exit $IER; fi
    done
}

# Private tools.
untar_private_tools () {
    FORCE="$1"
    PRIVATE_TOOL="$2"
    TARBALLS_DIR="$3"
    
    WORK_SUB_DIR="work"
    PARENT_DIR=".."
    
    check_bool "PRIVATE_TOOL" "$PRIVATE_TOOL"
    if test -d "${TARBALLS_DIR:?}"; then \
      (\cd "${TARBALLS_DIR:?}" 1> /dev/null && \
       MAKEFILE="Makefile_user" && \
       if ! test -f "$MAKEFILE"; then
         MAKEFILE="Makefile"; \
       fi && \
       TOOLS=`\
         $EGREP '^TOOLS_TO_INSTALL = ' "$MAKEFILE" | \
         $AWK -F' = ' '{print $2}'` && \
       LIST=`$ECHO "$TOOLS" | $STRIP_SPACE` && \
       if test -z "$LIST"; then return 0; fi && \
       $MKDIR "$WORK_SUB_DIR" && \
       (\cd "${WORK_SUB_DIR:?}" 1> /dev/null && \
        for TOOL in $TOOLS; do \
          TARBALL="$PARENT_DIR/$TOOL.tgz" && \
          if ! test -h "$TOOL" || "${PRIVATE_TOOL:?}" || "${FORCE:?}"; then \
            $RM $TOOL $TOOL-* && \
            $TARXZ "$TARBALL" && \
            TOOL_DIR=`$LD $TOOL-*` && \
            $LN "$TOOL_DIR" "$TOOL" && \
            untar_private_tools "$FORCE" true "$TOOL/tarballs"; \
          fi; \
        done)); \
    fi
}

remove_private_tools () {
    PROJECT_TARBALLS_DIR="$1"
    
    WORK_SUB_DIR="work"
    
    if test -d "${PROJECT_TARBALLS_DIR:?}"; then \
      (\cd "${PROJECT_TARBALLS_DIR:?}" 1> /dev/null && \
       $RM "$WORK_SUB_DIR"); \
    fi
}
