(***********************************************************************)
(*                                                                     *)
(*                              Ref-image                              *)
(*                                                                     *)
(*                  Image Segmentation by Refinement                   *)
(*                                                                     *)
(*               EPI Pomdapi, INRIA Paris - Rocquencourt               *)
(*                                                                     *)
(*  Copyright 2009-2013 INRIA.                                         *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Hend Ben Ameur <hbenameur@yahoo.ca>                                *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: scalar.ml,v 1.32 2013-12-02 15:28:38 fclement Exp $ *)

module Nat = Unatural;;

module Set = Umatrix_nonempty_set;;
module Mapping = Umatrix_graph;;
module Structural_mapping = Umatrix_structural_mapping;;

open Segmentation_types;;
open Image;;
open Zonation;;

module Make =
  functor (Of : Ordered_field_spec.ORDERED_FIELD) ->
struct
  include Of;;
  
  let char_is_zero = is_zero (of_int 128);;
  
  let _ppf = Configuration.get_formatter ();;
  
  (*let digest = Of.digest;;*)
  
  let sum = function
    | { Mapping.domain = dom; images = Mapping.Constant y; } ->
        of_nat (Set.cardinal dom) *. y
    | { Mapping.images = Mapping.Images ims; _; } ->
        Array.fold_left ( +. ) _0_ ims
  ;;
  
  let first_order_indicator g_zone (zt, zf) =
    let st = zt.Set.nonempty_set
    and sf = zf.Set.nonempty_set in
    let indict = sum (Mapping.restriction g_zone zt) in
    let indicf = sum (Mapping.restriction g_zone zf) in
    (Set.cardinal st, Set.cardinal sf,
     abs (indict -. indicf))
  ;;
  
  let exact_indicator ({ Mapping.domain = dom; _; } as g_zone) dicho =
    let (nt, nf, foi) = first_order_indicator g_zone dicho in
    let n = Set.cardinal dom in
    if Nat.add nt nf <> n then assert false else
    let ratio = of_nat n /. of_nat (Nat.mul nt nf) in
    inv_8_ *. ratio *. foi *. foi
  ;;
  
  let better_candidate_in_zone accu zcc =
    match accu with
    | No_more_zone_cutting -> Best_zone_cutting zcc
    | Best_zone_cutting best_zcc ->
        if indicator_of_zcc zcc <= indicator_of_zcc best_zcc then accu else
        Best_zone_cutting zcc
  ;;
  
  let gradient = sub_gimage ( -. );;
  
  let cut cs gdata zim =
    let gim = param zim in
    match gradient gdata gim with
    | { Mapping.images = Mapping.Constant g_y; _; }
        when char_is_zero g_y -> No_more_zone_cutting
    | ({ Mapping.images = Mapping.Constant _; _; } as g_gim)
    | ({ Mapping.images = Mapping.Images _; _; } as g_gim) ->
        let dom = g_gim.Mapping.domain in
        let dichos =
          match cs with
          | Overall_best ->
              let p idx _uidx = Mapping.image g_gim idx < _0_ in
              (*Printf.printf "Hello.\n";*)
              (*Format.fprintf ppf "Hello.\n";*)
              let dicho = Set.dichotomy p dom in
              [ dicho; ]
          | Best_in_a_family (nct, bcfl1) ->
              let bds = bounds_of_domain dom in
              Ulist.concat
                (Ulist1.to_list
                   (Ulist1.rev_map
                      (cuttings_from_family bds dom nct) bcfl1)) in
        let candidates =
          Ulist.map
            (fun dicho ->
               { indicator = exact_indicator g_gim dicho;
                 cutting = dicho; })
            dichos in
        Ulist.fold_left better_candidate_in_zone
          No_more_zone_cutting candidates
  ;;
  
  let mean gim = sum gim /. of_nat (Set.cardinal gim.Mapping.domain);;
  
  let sq_dist x y = square (x -. y);;
  let cost_sub gdata gim_sub =
    sum (sub_gimage sq_dist gdata gim_sub) /. _2_
  ;;
  
  let cost pim =
    let f accu _zn zc_gim =
      let { zc_cost = ct; _; } = Mapping.get_constant zc_gim in
      accu +. ct in
    pimage_fold_left_value f _0_ pim
  ;;
  
  let add_cost_decrease accu zim gdata dicho =
    let gim = param zim in
    let g_gim = gradient gdata gim in
    accu +. exact_indicator g_gim dicho
  ;;
end
;;

module Float = Make (Ordered_field.Float);;
