(***********************************************************************)
(*                                                                     *)
(*                              Ref-image                              *)
(*                                                                     *)
(*                  Image Segmentation by Refinement                   *)
(*                                                                     *)
(*               EPI Pomdapi, INRIA Paris - Rocquencourt               *)
(*                                                                     *)
(*  Copyright 2009-2013 INRIA.                                         *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Hend Ben Ameur <hbenameur@yahoo.ca>                                *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: state.ml,v 1.13 2013-12-02 15:28:38 fclement Exp $ *)

module Nat = Unatural;;
module Dgraph = Udynamic_graph;;

module Set = Umatrix_nonempty_set;;
module Mapping = Umatrix_graph;;
module Structural_mapping = Umatrix_structural_mapping;;
module Config = Configuration;;

let s_fun_name = Printf.sprintf "State.%s";;
let s_print_debug fun_name = Config.print_debug (s_fun_name fun_name);;

open Segmentation_types;;
open Segmentation_types_print;;
open Configuration_types;;
open Segmentation_spec;;
open Image;;
open Zonation;;

module Make =
  functor (Scal : SCALAR) ->
  functor (Cod : CODOMAIN) ->
struct
  type scalar = Scal.t;;
  type 'a codomain = 'a Cod.t;;
  
  type t = State of (scalar, other_components) zonation codomain
  and other_components = component_number list * t
  ;;
  
  let ppf = Config.get_formatter ();;
  
  let same (State zs1) (State zs2) = Cod.same Zonation.same zs1 zs2;;
  let print ppf (State zs) =
    Cod.print (print_zonation Scal.print (fun _ _ -> ())) ppf zs
  ;;
  let print_gimage = print_gimage (Cod.print Scal.print);;
  
  (*let digest = Cod.digest (Zonation.digest Scal.digest);;*)
  
  let gimage_components = function
    | { Mapping.domain = dom; images = Mapping.Constant y_tuple; } ->
        Cod.map (fun y_i -> Mapping.constant dom y_i) y_tuple
    | { Mapping.domain = dom; images = Mapping.Images tuple_ims; } ->
        let ims_tuple = Cod.array_components tuple_ims in
        Cod.map (fun ims_i -> Mapping.from_array dom ims_i) ims_tuple
  ;;
  
  let check_same_full fun_name dom gim =
    Set.check_same_full fun_name dom gim.Mapping.domain
  ;;
  
  let gimage_combine gims =
    let { Mapping.domain = dom0; _; } = Cod.get gims Nat._0_ in
    Cod.iter (check_same_full (s_fun_name "gimage_combine") dom0) gims;
    let f idx _uidx = Cod.map (fun gim -> Mapping.image gim idx) gims in
    Mapping.from_fun dom0 f
  ;;
  
  let gimage_datas (State zs) = Cod.map (fun z -> z.gimage_data) zs;;
  
  let cimages (State zs) = Cod.map (fun z -> z.cimage) zs;;
  
  let cost (State zs) =
    Cod.fold_left
      (fun accu { cimage = (j, _); _; } -> Scal.( +. ) accu j)
      Scal._0_ zs
  ;;
  
  let cut cs dm gdata (other_channell, State zs) zn map =
    let bzc = Scal.cut cs gdata map in
    match dm with
    | (Best_cutting, All_components) ->
        begin match bzc with
        | No_more_zone_cutting -> bzc
        | Best_zone_cutting
              ({ indicator = delta_j; cutting = dicho; } as zcc) ->
            let add_other_channel_delta_j accu other_channel =
              let { gimage_data = gdata;
                    cimage = (_, pim); _; } = Cod.get zs other_channel in
              let { Dgraph.value = gim; _; } =
                Structural_mapping.get pim zn in
              Scal.add_cost_decrease accu gim gdata dicho in
            let delta_jt =
              Ulist.fold_left add_other_channel_delta_j
                delta_j other_channell in
            Best_zone_cutting { zcc with indicator = delta_jt; } end
    | (Best_cutting, Best_component)
    | (All_cuttings, Best_component)
    | (All_cuttings, All_components) -> bzc
  ;;
  
  let init () = (* Only used for cold initialization. *)
    let idata = Options.get_argument_initial_image () in
    assert (s_print_debug "init" "idata" print_image ppf idata);
    let gdata = gimage_of_image (Cod.of_color Scal.of_int) idata in
    assert (s_print_debug "init" "gdata" print_gimage ppf gdata);
    let gdatas = gimage_components gdata in
    let cs = Config.get_cutting_strategy () in
    let dm = Config.get_dimension_management () in
    let make_trivial_zonation gdata =
      make_trivial_zonation
        (cut cs dm gdata) Scal.mean Scal.cost_sub Scal.cost gdata in
    let channels = Cod.all_channels in
    let other_channells = Cod.complementary channels in
    let zs =
      Cod.map3
        make_trivial_zonation
        gdatas channels other_channells in
    State zs
  ;;
  
  let save_final_image dname gim_n =
    let im_n = image_of_full_gimage (Cod.to_color Scal.to_int) gim_n in
    let fname = Config.get_final_image_file_name () in
    Ppm.save_pixmap (Filename.concat dname fname) (Umatrix.to_array_array im_n)
  ;;
  
  let final dname (State zs_n) =
    let gims_n = Cod.map gimage_of_zonation zs_n in
    let gim_n = gimage_combine gims_n in
    save_final_image dname gim_n
  ;;
end
;;
