(***********************************************************************)
(*                                                                     *)
(*                                Ulib                                 *)
(*                                                                     *)
(*                   OCaml user's additional library                   *)
(*                                                                     *)
(*               EPI Pomdapi, INRIA Paris - Rocquencourt               *)
(*                                                                     *)
(*  Copyright 2004-2014 INRIA.                                         *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: udynamic_array.ml,v 1.5 2014/01/16 13:35:38 fclement Exp $ *)

(* Dynamic arrays. *)

let da_failwith = Upervasives.failwith "Udynamic_array";;

type index = Unatural.t;;
type length = Unatural.t;;

type 'a t =
  {
   mutable array : 'a array;
   default : 'a;
  }
;;

let same da1 da2 = da1 == da2;;

let max_length = Unatural.of_int (Sys.max_array_length / 2);;

let length da = Uarray.length da.array;;

let make n x = { array = Uarray.make n x; default = x; };;

let get da i = Array.get da.array (Unatural.to_int i);;

let grow fun_name da i =
  let len = length da in
  if i >= len then
    let new_len = max (Unatural.succ i) (Unatural.mul Unatural._2_ len) in
    let new_len = if new_len > max_length then max_length else new_len in
    if new_len < i then
      da_failwith fun_name "cannot create such a large dynamic array" else
    let new_a = Uarray.make new_len da.default in
    Array.blit da.array 0 new_a 0 (Unatural.to_int len);
    da.array <- new_a
;;

let set da i x =
  grow "set" da i;
  da.array.(Unatural.to_int i) <- x
;;

(* Converters. *)

let to_array da = da.array;;
let of_array x a = { array = a; default = x; };;

(* Iterators. *)

let iter f da = Array.iteri (fun i -> f (Unatural.of_int i)) da.array;;

let map f da =
  {
   array = Array.mapi (fun i -> f (Unatural.of_int i)) da.array;
   default = f Unatural._0_ da.default;
  }
;;

let fold_left f accu da =
  let ff accu i = f accu (Unatural.of_int i) in
  Uarray.fold_left_i ff accu da.array
;;
