(***********************************************************************)
(*                                                                     *)
(*                                Ulib                                 *)
(*                                                                     *)
(*                   OCaml user's additional library                   *)
(*                                                                     *)
(*               EPI Pomdapi, INRIA Paris - Rocquencourt               *)
(*                                                                     *)
(*  Copyright 2004-2014 INRIA.                                         *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: ulistn_spec.mli,v 1.10 2014/01/17 19:19:30 fclement Exp $ *)

(** {3 Module types for lists with minimum length} *)

module type LISTN_STRUCTURE =
sig
  (** {3 Lists with minimum length} *)
  
  type 'a listpn;;
  type 'a listn;;
  type 'a t = 'a listn;;
  (** The type of lists with minimum length. *)
  
  val min_length : Unatural.t;;
  (** Minimum length of the lists. *)
  
  val to_listpn : 'a listn -> 'a listpn;;
  (** [to_listpn ln] returns the elements of [ln] as a list with minimum
   length the positive part of [pred] {!min_length}. *)
  
  val of_listpn : 'a listpn -> 'a listn;;
  (** [of_listpn lpn] returns the elements of [lpn], a list with minimum
   length the positive part of [pred] {!min_length}, as a list with minimum
   length {!min_length}.
   
   Raise [Invalid_argument] if the list is not long enough. *)
  
  val to_list : 'a listn -> 'a list;;
  (** [to_list ln] returns the (usual) list of all the elements of [ln]. *)
  
  val of_list : 'a list -> 'a listn;;
  (** [of_list l] returns the list with minimum length containing the elements
   of [l].
   
   Raise [Invalid_argument] if the list is not long enough. *)
  
  val cons : 'a -> 'a listn -> 'a listn;;
  (** [cons h ln] places [h] in front of [ln]. *)
  
  val hd : 'a listn -> 'a;;
  (** Return the first element of the given list with minimum length. *)
  
  val tl : 'a listn -> 'a listpn;;
  (** Return the given list with minimum length without its first element. *)
end
;;

module type LISTN =
sig
  (** {3 Operations on lists with minimum length} *)
  
  include LISTN_STRUCTURE;;
  
  val lift : ('a list -> 'b list) -> 'a listn -> 'b listn;;
  (** Lift an operation on lists into an operation on lists with minimum
   length. *)
  val lift2 :
    ('a list -> 'b list -> 'c list) -> 'a listn -> 'b listn -> 'c listn
  ;;
  (** Same as {!lift} for binary operations. *)
  
  val length : 'a listn -> Unatural.t;;
  (** Return the length (number of elements) of the given list with minimum
   length. *)
  
  val first : int -> 'a listn -> 'a list;;
  (** [first p ln] returns the [p] first elements of [ln].
   If [{!length} ln <= p] then [first p ln] returns [{!to_list} ln]. *)
  
  val last : int -> 'a listn -> 'a list;;
  (** [last p ln] returns the [p] last elements of [ln].
   If [{!length} ln <= p] then [last p ln] returns [{!to_list} ln]. *)
  
  val nth : 'a listn -> int -> 'a;;
  (** [nth ln n] returns the [n]-th element of [ln].
   The first element (head of the list) is at position 0.
   
   Raise [Failure] if the list with minimum length is too short.
   Raise [Invalid_argument] if the integer argument is negative. *)
  
  val rank : 'a -> 'a listn -> int;;
  (** [rank e l] returns [r] such that {!nth} [l r] = [e]. *)
  
  val rev : 'a listn -> 'a listn;;
  (** List with minimum length reversal. *)
  
  val append : 'a listn -> 'a listn -> 'a listn;;
  (** Catenate two lists with minimum length.
   Not tail-recursive (length of the first argument). *)
  
  val rev_append : 'a listn -> 'a listn -> 'a listn;;
  (** [rev_append ln1 ln2] is equivalent to {!append} [(]{!rev} [ln1) ln2],
   but is tail-recursive and more efficient. *)
  
  val concat : 'a listn listn -> 'a listn;;
  (** Concatenate a list with minimum length of lists with minimum length.
   The elements of the argument are all concatenated together (in the same
   order) to give the result.
   Tail-recursive. *)
  
  val flatten : 'a listn listn -> 'a listn;;
  (** Alias for {!concat}. *)
  
  val init : Unatural.t -> (int -> 'a) -> 'a listn;;
  (** [init n f] is equivalent to
   {!LISTN_STRUCTURE.of_list} [(Ulist.init n f)].
   Thus [n] is the length of the resulting list with minimum length.
   
   Raise [Invalid_argument] is [n] is lower than
   {!LISTN_STRUCTURE.min_length}. *)
  
  val rev_init : Unatural.t -> (int -> 'a) -> 'a listn;;
  (** [rev_init n f] is equivalent to {!rev} [(]{!init} [n f]. *)
  
  val replace : int -> 'a -> 'a listn -> 'a listn;;
  (** [replace i x ln] replaces [i]-th element of [ln] by [x]. *)
  
  val rev_replace : int -> 'a -> 'a listn -> 'a listn;;
  (** [rev_replace i x ln] is equivalent to {!rev} [(]{!replace} [i x ln)],
   but is tail-recursive and more efficient. *)
  
  val replace_by_list : int -> 'a listn -> 'a listn -> 'a listn;;
  (** [replace_by_list i ln1 ln2] replaces the [i]-th element of [ln2] by the
   elements of [ln1] (in the same order).
   Not tail-recursive. *)
  
  val rev_replace_by_list : int -> 'a listn -> 'a listn -> 'a listn;;
  (** [rev_replace_by_list i ln1 ln2] is equivalent to
   {!replace_by_list} [i (]{!rev} [ln1) ln2],
   but is tail-recursive and more efficient. *)
  
  val insert : int -> 'a listn -> 'a listn -> 'a listn;;
  (** [insert i ln1 ln2] inserts the elements of [ln1] (in the same order)
   right before the [i]-th element of [ln2].
   Hence, [insert 0] is equivalent to {!append}. *)
  
  val rev_insert : int -> 'a listn -> 'a listn -> 'a listn;;
  (** [rev_insert i ln1 ln2] is equivalent to {!rev} [(]{!insert} [i ln1 ln2)],
   but is tail-recursive and more efficient. *)
  
  (** {6 Iterators} *)
  
  val iter : ('a -> unit) -> 'a listn -> unit;;
  (** [iter f ln] applies [f] in turn to all elements of [ln]. *)
  
  val iter_i : (int -> 'a -> unit) -> 'a listn -> unit;;
  (** Same as {!iter} but the function is applied to the index of the element
   as first additional argument (counting from 0). *)
  
  val map : ('a -> 'b) -> 'a listn -> 'b listn;;
  (** [map f ln] applies [f] in turn to all elements of [ln], and returns the
   list with minimum length of results.
   Not tail-recursive. *)
  
  val map_i : (int -> 'a -> 'b) -> 'a listn -> 'b listn;;
  (** Same as {!map} but the function is applied to the index of the element
   as first additional argument (counting from 0). *)
  
  val rev_map : ('a -> 'b) -> 'a listn -> 'b listn;;
  (** [rev_map f l] is equivalent to {!rev} [(]{!map} [f l)],
   but is tail-recursive and more efficient. *)
  
  val rev_map_i : (int -> 'a -> 'b) -> 'a listn -> 'b listn;;
  (** Same as {!rev_map} but the function is applied to the index of the
   element as first additional argument (counting from 0). *)
  
  val flat_map : ('a -> 'b listn) -> 'a listn -> 'b listn;;
  (** [flat_map f l] is equivalent to {!flatten} [(]{!map} [f l)].
   Tail-recursive. *)
  
  val flat_map_i : (int -> 'a -> 'b listn) -> 'a listn -> 'b listn;;
  (** Same as {!flat_map} but the function is applied to the index of the
   element as first additional argument (counting from 0).
   Tail-recursive. *)
  
  val rev_flat_map : ('a -> 'b listn) -> 'a listn -> 'b listn;;
  (** [rev_flat_map f l] is equivalent to {!rev} [(]{!flat_map} [f l)].
   Tail-recursive. *)
  
  val rev_flat_map_i : (int -> 'a -> 'b listn) -> 'a listn -> 'b listn;;
  (** Same as {!rev_flat_map} but the function is applied to the index of the
   element as first additional argument (counting from 0).
   Tail-recursive. *)
  
  val fold_left : ('a -> 'b -> 'a) -> 'a -> 'b listn -> 'a;;
  (** [fold_left f a (b0, lpn)] is equivalent to
   [Listpn.fold_left f (f a b0) lpn] where [Listpn] is the module
   implementing lists with minimum length being the positive part of
   [pred] {!LISTN_STRUCTURE.min_length}.
   Tail-recursive. *)
  
  val fold_left_i : (int -> 'a -> 'b -> 'a) -> 'a -> 'b listn -> 'a;;
  (** Same as {!fold_left}, but the function is applied to the index of the
   element as first additional argument (counting from 0).
   Tail-recursive. *)
  
  val fold_right : ('a -> 'b -> 'b) -> 'a listn -> 'b -> 'b;;
  (** [fold_right f (a1, lpn) b] is equivalent to
   [f a1 (Listpn.fold_left f lpn b)] where [Listpn] is the module
   implementing lists with minimum length being the positive part of
   [pred] {!LISTN_STRUCTURE.min_length}.
   Not tail-recursive. *)
  
  val fold_right_i : (int -> 'a -> 'b -> 'b) -> 'a listn -> 'b -> 'b;;
  (** Same as {!fold_right}, but the function is applied to the index of the
   element as first additional argument (counting from 0).
   Not tail-recursive. *)
  
  val rev_fold_left_env :
    ('env * 'b list -> 'a -> 'env * 'b listn) ->
    'env * 'b list -> 'a listn -> 'env * 'b listn
  ;;
  (** [rev_fold_left_env] is an alternative of {!fold_left} dedicated to
   accumulators of type ['env * 'b list].
   Here, we can further ensure that the resulting list has minimum length
   {!LISTN_STRUCTURE.min_length}. *)
  
  val fold_left_env_init :
    ('env * 'b list -> 'a -> 'env * 'b listn) ->
    'env -> 'a listn -> 'env * 'b listn
  ;;
  (** [fold_left_env_init] is the partial application of {!rev_fold_left_env}
   to an initial accumulator [(env, [])].
   In addition, [fold_left_env_init] preserves the order of the list arguments
   in the final result.
   
   More precisely, [fold_left_env_init f env l] is equivalent to
   [let (env, accu) = rev_fold_left_env f (env, []) l in
    (env, rev accu)]. *)
  
  (** {6 Iterators on two nonempty lists} *)
  val iter2 : ('a -> 'b -> unit) -> 'a listn -> 'b listn -> unit;;
  val iter2_i : (int -> 'a -> 'b -> unit) -> 'a listn -> 'b listn -> unit;;
  val map2 : ('a -> 'b -> 'c) -> 'a listn -> 'b listn -> 'c listn;;
  val map2_i : (int -> 'a -> 'b -> 'c) -> 'a listn -> 'b listn -> 'c listn;;
  val rev_map2 : ('a -> 'b -> 'c) -> 'a listn -> 'b listn -> 'c listn;;
  val rev_map2_i : (int -> 'a -> 'b -> 'c) -> 'a listn -> 'b listn -> 'c listn;;
  val flat_map2 : ('a -> 'b -> 'c listn) -> 'a listn -> 'b listn -> 'c listn;;
  val flat_map2_i :
    (int -> 'a -> 'b -> 'c listn) -> 'a listn -> 'b listn -> 'c listn
  ;;
  val rev_flat_map2 :
    ('a -> 'b -> 'c listn) -> 'a listn -> 'b listn -> 'c listn
  ;;
  val rev_flat_map2_i :
    (int -> 'a -> 'b -> 'c listn) -> 'a listn -> 'b listn -> 'c listn
  ;;
  val fold_left2 : ('a -> 'b -> 'c -> 'a) -> 'a -> 'b listn -> 'c listn -> 'a;;
  val fold_left2_i :
    (int -> 'a -> 'b -> 'c -> 'a) -> 'a -> 'b listn -> 'c listn -> 'a
  ;;
  val fold_right2 : ('a -> 'b -> 'c -> 'c) -> 'a listn -> 'b listn -> 'c -> 'c;;
  val fold_right2_i :
    (int -> 'a -> 'b -> 'c -> 'c) -> 'a listn -> 'b listn -> 'c -> 'c
  ;;
  val rev_fold_left2_env :
    ('env * 'c list -> 'a -> 'b -> 'env * 'c listn) ->
    'env * 'c list -> 'a listn -> 'b listn -> 'env * 'c listn
  ;;
  val fold_left2_env_init :
    ('env * 'c list -> 'a -> 'b -> 'env * 'c listn) ->
    'env -> 'a listn -> 'b listn -> 'env * 'c listn
  ;;
  (** Same as {!iter}, {!iter_i}, {!map}, {!map_i}, {!rev_map}, {!rev_map_i},
   {!flat_map}, {!flat_map_i}, {!rev_flat_map}, {!rev_flat_map_i},
   {!fold_left}, {!fold_left_i}, {!fold_right}, {!fold_right_i},
   {!rev_fold_left_env}, and {!fold_left_env_init},
   but the argument function takes an element of the second list with minimum
   length as an extra argument, right after the argument element of the first
   list with minimum length.
   
   Raise [Invalid_argument] if the two lists with minimum length have different
   lengths. *)
  
  (** {6 Scanning} *)
  
  val for_all : ('a -> bool) -> 'a listn -> bool;;
  (** Check if all elements of the list with minimum length satisfy the given
   predicate. *)
  
  val exists : ('a -> bool) -> 'a listn -> bool;;
  (** Check if at least one element of the list with minimum length satisfies
   the given predicate. *)
  
  val for_all2 : ('a -> 'b -> bool) -> 'a listn -> 'b listn -> bool;;
  (** Same as {!for_all}, but for a two-argument predicate.
   
   Raise [Invalid_argument] if the two lists have different lengths. *)
  
  val exists2 : ('a -> 'b -> bool) -> 'a listn -> 'b listn -> bool;;
  (** Same as {!exists}, but for a two-argument predicate.
   
   Raise [Invalid_argument] if the two lists have different lengths. *)
  
  val mem : 'a -> 'a listn -> bool;;
  (** [mem a ln] is true if and only if [a] is equal to an element of [ln]. *)
  
  val memq : 'a -> 'a listn -> bool;;
  (** Same as {!mem}, but uses physical equality instead of structural equality
   to compare list elements. *)
  
  (** {6 Searching} *)
  
  val find : ('a -> bool) -> 'a listn -> 'a;;
  (** [find p ln] returns the first element of list with minimum length [ln]
   that satisfies predicate [p].
   
   Raise [Not_found] if there is no such value. *)
  
  val filter : ('a -> bool) -> 'a listn -> 'a list;;
  (** [filter p ln] returns all the elements of list with minimum length [ln]
   that satisfy predicate [p].
   Order of elements in the input list is preserved.  *)
  
  val find_all : ('a -> bool) -> 'a listn -> 'a list;;
  (** Alias for {!filter}. *)
  
  val partition : ('a -> bool) -> 'a listn -> 'a list * 'a list;;
  (** [partition p ln] returns a pair of lists [(l1, l2)], where [l1] is the
   list of all elements of [l] that satisfy predicate [p], and [l2] is the list
   of all elements of [l] that do not satisfy [p].
   Order of elements in the input list is preserved. *)
  
  (** {6 Association lists with minimum length} *)
  
  val assoc : 'a -> ('a * 'b) listn -> 'b;;
  (** [assoc a ln] returns the first value associated with key [a] in list of
   pairs with minimum length [ln].
   
   Raise [Not_found] if there is no such value. *)
  
  val assq : 'a -> ('a * 'b) listn -> 'b;;
  (** Same as {!assoc}, but uses physical equality instead of structural
   equality to compare keys. *)
  
  val mem_assoc : 'a -> ('a * 'b) listn -> bool;;
  (** Same as {!assoc}, but simply return true if a binding exists, and false
   if no bindings exist for the given key. *)
  
  val mem_assq : 'a -> ('a * 'b) listn -> bool;;
  (** Same as {!mem_assoc}, but uses physical equality instead of structural
   equality to compare keys. *)
  
  val remove_assoc : 'a -> ('a * 'b) listn -> ('a * 'b) list;;
  (** [remove_assoc a ln] returns the list of pairs [l] without the first pair
   with key [a], if any.
   Not tail-recursive. *)
  
  val remove_assq : 'a -> ('a * 'b) listn -> ('a * 'b) list;;
  (** Same as {!remove_assoc}, but uses physical equality instead of structural
   equality to compare keys.
   Not tail-recursive. *)
  
  val compare_assoc : ('a -> 'a -> int) -> ('a * 'b) listn -> 'a * 'b;;
  (** [compare_assoc cmp ln] returns the first element of list of pairs with
   minimum length [ln] with the minimum first component according to comparison
   function [cmp]. *)
  
  val min_assoc : ('a * 'b) listn -> 'a * 'b;;
  (** [min_assoc] is equivalent to {!compare_assoc} [Pervasives.compare]. *)
  
  val max_assoc : ('a * 'b) listn -> 'a * 'b;;
  (** [max_assoc] is equivalent to
   {!compare_assoc} [(fun x y -> - (Pervasives.compare x y))]. *)
  
  val compare_fst_assoc : ('a -> 'a -> int) -> ('a * 'b) listn -> 'a * 'b;;
  (** Alias for {!compare_assoc}. *)
  
  val min_fst_assoc : ('a * 'b) listn -> 'a * 'b;;
  (** Alias for {!min_assoc}. *)
  
  val max_fst_assoc : ('a * 'b) listn -> 'a * 'b;;
  (** Alias for {!max_assoc}. *)
  
  val compare_snd_assoc : ('a -> 'a -> int) -> ('b * 'a) listn -> 'b * 'a;;
  (** Same as {!compare_assoc} when comparing second components. *)
  
  val min_snd_assoc : ('a * 'b) listn -> 'a * 'b;;
  (** Same as {!min_assoc} when comparing second components. *)
  
  val max_snd_assoc : ('a * 'b) listn -> 'a * 'b;;
  (** Same as {!min_assoc} when comparing second components. *)
  
  (** {6 Lists of pairs with minimum length} *)
  
  val split : ('a * 'b) listn -> 'a listn * 'b listn;;
  (** Transform a list of pairs with minimum length into a pair of lists with
   minimum length.
   Not tail-recursive. *)
  
  val combine : 'a listn -> 'b listn -> ('a * 'b) listn
  (** Transform a pair of lists with minimum length into a list of pairs with
   minimum length.
   Not tail-recursive.
   
   Raise [Invalid_argument] if the two lists have different lengths. *)
  
  (** {6 Sorting} *)
  
  val sort : ('a -> 'a -> int) -> 'a listn -> 'a listn;;
  (** Sort a list with minimum length in increasing order according to a
   comparison function.
   The comparison function must behave the same as [Pervasives.compare]
   (see [Array.sort] for a complete specification).
   Run in constant heap space and logarithmic stack space. *)
  
  val stable_sort : ('a -> 'a -> int) -> 'a listn -> 'a listn;;
  (** Same as {!sort}, but the sorting algorithm is stable (i.e. elements that
   compare equal are kept in their original order). *)
  
  val fast_sort : ('a -> 'a -> int) -> 'a listn -> 'a listn;;
  (** Same as {!sort} or {!stable_sort}, whichever is faster on typical
   input. *)
  
  val merge : ('a -> 'a -> int) -> 'a listn -> 'a listn -> 'a listn;;
  (** Merge two lists with minimum length.
   Assuming that [ln1] and [ln2] are sorted according to comparison function
   [cmp], [merge cmp ln1 ln2] returns a sorted list containing all the elements
   of [ln1] and [ln2].
   If several elements compare equal, the elements of [ln1] are listed before
   the elements of [ln2].
   Not tail-recursive (sum of the lengths of the arguments). *)

  val of_range : ('a -> 'a) -> 'a -> 'a -> 'a listn;;
  (** [of_range pred l r] returns
   [of_list \[ l; ...; pred (pred r); pred r; \]]. *)
  
  (**/**)
  (** {6 Undocumented functions} *)
  val iter_i_shift : (int -> 'a -> unit) -> int -> 'a listn -> unit;;
  val map_i_shift : (int -> 'a -> 'b) -> int -> 'a listn -> 'b listn;;
  val rev_map_i_end_shift :
      'b list -> (int -> 'a -> 'b) -> int -> 'a listn -> 'b listn
  ;;
  val flat_map_end : 'b list -> ('a -> 'b listn) -> 'a listn -> 'b listn;;
  val rev_flat_map_end : 'b list -> ('a -> 'b listn) -> 'a listn -> 'b listn;;
  val flat_map_i_end_shift :
    'b list -> (int -> 'a -> 'b listn) -> int -> 'a listn -> 'b listn
  ;;
  val rev_flat_map_i_end_shift :
    'b list -> (int -> 'a -> 'b listn) -> int -> 'a listn -> 'b listn
  ;;
  val fold_left_i_shift :
    (int -> 'a -> 'b -> 'a) -> int -> 'a -> 'b listn -> 'a
  ;;
  val fold_right_i_shift :
    (int -> 'a -> 'b -> 'b) -> int -> 'a listn -> 'b -> 'b
  ;;
  val fold_left_env :
    ('env * 'b list -> 'a -> 'env * 'b listn) ->
    'env * 'b list -> 'a listn -> 'env * 'b listn
  ;;
  
  val iter2_i_shift :
    (int -> 'a -> 'b -> unit) -> int -> 'a listn -> 'b listn -> unit
  ;;
  val map2_i_shift :
    (int -> 'a -> 'b -> 'c) -> int -> 'a listn -> 'b listn -> 'c listn
  ;;
  val rev_map2_i_end_shift :
    'c list -> (int -> 'a -> 'b -> 'c) ->
    int -> 'a listn -> 'b listn -> 'c listn
  ;;
  val flat_map2_end :
    'c list -> ('a -> 'b -> 'c listn) -> 'a listn -> 'b listn -> 'c listn
  ;;
  val rev_flat_map2_end :
    'c list -> ('a -> 'b -> 'c listn) -> 'a listn -> 'b listn -> 'c listn
  ;;
  val flat_map2_i_end_shift :
    'c list -> (int -> 'a -> 'b -> 'c listn) ->
    int -> 'a listn -> 'b listn -> 'c listn
  ;;
  val rev_flat_map2_i_end_shift :
    'c list -> (int -> 'a -> 'b -> 'c listn) ->
    int -> 'a listn -> 'b listn -> 'c listn
  ;;
  val fold_left2_i_shift :
    (int -> 'a -> 'b -> 'c -> 'a) -> int -> 'a -> 'b listn -> 'c listn -> 'a
  ;;
  val fold_right2_i_shift :
    (int -> 'a -> 'b -> 'c -> 'c) -> int -> 'a listn -> 'b listn -> 'c -> 'c
  ;;
  val fold_left2_env :
    ('env * 'c list -> 'a -> 'b -> 'env * 'c listn) ->
    'env * 'c list -> 'a listn -> 'b listn -> 'env * 'c listn
  ;;
end
;;

module type LISTN_SUCC =
  functor (Listn : LISTN) ->
  LISTN with
  type 'a listpn = 'a Listn.t and
  type 'a listn = 'a * 'a Listn.t
;;
