(***********************************************************************)
(*                                                                     *)
(*                                Ulib                                 *)
(*                                                                     *)
(*                   OCaml user's additional library                   *)
(*                                                                     *)
(*               EPI Pomdapi, INRIA Paris - Rocquencourt               *)
(*                                                                     *)
(*  Copyright 2004-2014 INRIA.                                         *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: umatrix.mli,v 1.24 2014/01/10 15:28:04 fclement Exp $ *)

(** {3 Matrices}
   
   A matrix is a two-dimensional array, or equivalently an array of values
   such that each value is an array of a fixed length. An element in a matrix
   is indexed by a pair of integers and accessed in first-index-major order,
   that is, " la C" or using the lexicographic ordering for pairs of integers.
   
   Following the classical mathematical convention, the first index to access
   an element refers to the row number and is denoted by [i], and conversely
   the second index refers to the column number and is denoted by [j]. As
   usual, the row numbers are increasing from top to bottom and the column
   numbers are increasing from left to right.
   
   Beware: these notions are conventional and subject to modification
   according to the problem at hand. For instance, in graphical applications
   the row numbers may decrease from top to bottom. *)

type 'a matrix = 'a Umatrix_types.matrix;;
type 'a t = 'a matrix;;
(** The type of matrices. *)

val of_array_array : 'a array array -> 'a matrix;;
(** Convert the given array of arrays into a matrix. *)

val to_array_array : 'a matrix -> 'a array array;;
(** Convert the given matrix into an array of arrays. *)

val size : 'a matrix -> Unatural.t * Unatural.t;;
(** Return the size of the given matrix. The size of a matrix is a pair of
   naturals that gives the number of elements in both directions.
   
   Raise [Invalid_argument] if the given array is not a valid matrix, that is,
   if it is not an array of arrays that all have the same length. *)

val length : 'a matrix -> Unatural.t * Unatural.t;;
(** Alias for {!size}. *)

val dim_i : 'a matrix -> Unatural.t;;
(** [dim_i m] returns the first dimension of the matrix [m]. Hence, the first
   index of [m] varies from [0] to {!Unatural.to_int} [(dim_i m) - 1]. *)

val dim_j : 'a matrix -> Unatural.t;;
(** [dim_j m] returns the second dimension of the matrix [m]. Hence, the
   second index of [m] varies from [0] to {!Unatural.to_int} [(dim_j m) - 1]. *)

val dim1 : 'a matrix -> Unatural.t;;
(** Alias for {!dim_i}. *)

val dim2 : 'a matrix -> Unatural.t;;
(** Alias for {!dim_j}. *)

val get : 'a matrix -> int -> int -> 'a;;
(** [get m i j] returns the element number [(i, j)] of matrix [m].
   
   Raise [Invalid_argument] if [i] is outside the range [0] to {!dim_i} [m - 1]
   or [j] is outside the range [0] to {!dim_j} [m - 1]. *)

val set : 'a matrix -> int -> int -> 'a -> unit;;
(** [set m i j x] modifies matrix [m] in place, replacing element number
   [(i, j)] with [x].
   
   Raise [Invalid_argument], if [i] is outside the range [0] to
   {!dim_i} [m - 1] or [j] is outside the range [0] to {!dim_j} [m - 1]. *)

val make : Unatural.t -> Unatural.t -> 'a -> 'a matrix;;
(** [make ni nj x] returns a fresh matrix with size [(ni, nj)]. All the
   elements of this new matrix are initially physically equal to [x].
   
   Raise [Invalid_argument], if {!Unatural.to_int} [ni] or
   {!Unatural.to_int} [nj] is greater than [Sys.max_array_length]. If the
   value of [x] is a floating-point number, then the maximum size is only
   [Sys.max_array_length / 2]. *)

val init : Unatural.t -> Unatural.t -> (int -> int -> 'a) -> 'a matrix;;
(** [init ni nj f] returns a fresh matrix of size [(ni, nj)], with element
   number [(i, j)] initialized to the result of [f i j]. In other terms, it
   tabulates the results of [f] applied to the pairs of integers [(0, 0)] to
   [(]{!Unatural.to_int} [ni - 1,] {!Unatural.to_int} [nj - 1)].
   
   Raise [Invalid_argument], if {!Unatural.to_int} [ni] or
   {!Unatural.to_int} [nj] is greater than [Sys.max_array_length]. If the
   return type of [f] is [float], then the maximum size is only
   [Sys.max_array_length / 2]. *)

val append_i : 'a matrix -> 'a matrix -> 'a matrix;;
(** [append_i m1 m2] returns a fresh matrix containing the concatenation of
   matrices [m1] and [m2] along the first direction.
   
   Raise [Invalid_argument], if the two matrices have different second
   dimensions. *)

val append_j : 'a matrix -> 'a matrix -> 'a matrix;;
(** [append_j m1 m2] returns a fresh matrix containing the concatenation of
   matrices [m1] and [m2] along the second direction.
   
   Raise [Invalid_argument], if the two matrices have different first
   dimensions. *)

val append1 : 'a matrix -> 'a matrix -> 'a matrix;;
(** Alias for {!append_i}. *)

val append2 : 'a matrix -> 'a matrix -> 'a matrix;;
(** Alias for {!append_j}. *)

val append : 'a matrix -> 'a matrix -> 'a matrix -> 'a matrix -> 'a matrix;;
(** [append m1 m2 m3 m4] returns a fresh matrix containing the concatenation
   of matrices [m1], [m2]. [m3] and [m4]. It returns the same result as
   {!append_i} [(]{!append_j} [m1 m2) (]{!append_j} [m3 m4)]. *)

val concat_i : 'a matrix list -> 'a matrix;;
(** Same as {!append_i}, but concatenates a list of matrices. *)

val concat_j : 'a matrix list -> 'a matrix;;
(** Same as {!append_j}, but concatenates a list of matrices. *)

val concat1 : 'a matrix list -> 'a matrix;;
(** Alias for {!concat_i}. *)

val concat2 : 'a matrix list -> 'a matrix;;
(** Alias for {!concat_j}. *)

val concat : 'a matrix list list -> 'a matrix;;
(** [concat mll] returns a fresh matrix containing the concatenation of
   matrices of the list of lists [mll]. *)

val sub : 'a matrix -> int -> Unatural.t -> int -> Unatural.t -> 'a matrix;;
(** [sub m oi li oj lj] returns a fresh matrix of size [(li, lj)], containing
   the elements number [(oi, oj)] to [(oi +] {!Unatural.to_int} [li - 1,
   oj +] {!Unatural.to_int} [lj - 1)] of matrix [m].
   
   Raise [Invalid_argument], if [oi], [li], [oj], and [lj] do not designate a
   valid submatrix of [m]; that is, if [oi < 0], or
   [oi +] {!Unatural.to_int} [li >] {!dim_i} [m], or [oj < 0], or
   [oj +] {!Unatural.to_int} [lj >] {!dim_j} [m]. *)

val copy : 'a matrix -> 'a matrix;;
(** [copy m] returns a copy of [m], that is, a fresh matrix containing the
   same elements as [m]. *)

val fill : 'a matrix -> int -> Unatural.t -> int -> Unatural.t -> 'a -> unit;;
(** [fill m oi li oj lj x] modifies the matrix [m] in place, storing [x] in
   elements number [(oi, oj)] to [(oi +] {!Unatural.to_int} [li - 1, oj +]
   {!Unatural.to_int} [lj - 1)].
   
   Raise [Invalid_argument], if [oi], [li], [oj] and [lj] do not designate a
   valid submatrix of [m] (see {!sub}). *)

val blit :
  'a matrix -> int -> int -> 'a matrix -> int -> int ->
  Unatural.t -> Unatural.t -> unit
;;
(** [blit m1 oi1 oj1 m2 oi2 oj2 li lj] copies {!Unatural.to_int} [li *]
   {!Unatural.to_int} [lj] elements from matrix [m1], starting at element number
   [(oi1, oj1)], to matrix [m2], starting at element number [(oi2, oj2)]. It
   works correctly even if [m1] and [m2] are the same matrix, and the source
   and destination chunks overlap.
   
   Raise [Invalid_argument], if [oi1], [li], [oj1] and [lj] do not designate
   a valid submatrix of [m1], or if [oi2], [li], [oj2] and [lj] do not
   designate a valid submatrix of [m2] (see {!sub}). *)

val transpose : 'a matrix -> 'a matrix;;
(** [transpose m] returns the transposed of matrix [m], that is, a fresh matrix
   of size [(]{!dim_j} [m,] {!dim_i} [m)] whose element number [(i, j)] is
   equal to [m.(j).(i)].
   
   Transposition is an involutive operation, except for matrices with at least
   one zero dimension, in which case the transposed matrix is the empty matrix
   [[||]] (of size [(]{!Unatural._0_}[,] {!Unatural._0_}[)]). *)

val to_array : 'a matrix -> 'a array;;
(** [to_array m] returns a fresh array containing the concatenation of all
   subarrays along the first direction of matrix [m]. It is equivalent to
   [Array.concat [| m.(0); m.(1); ...; m.(]{!Unatural.to_int} [ni - 1) |]] where
   [ni] is the first dimension of matrix [m]. *)

val of_array : 'a array -> Unatural.t -> Unatural.t -> 'a matrix;;
(** [of_array a ni nj] returns a fresh matrix containing the split of array
   [a] into [ni] subarrays of length [nj].
   
   Raise [Invalid_argument], if [Array.length a <>] {!Unatural.to_int} [ni *]
   {!Unatural.to_int} [nj]. *)

val to_arraylist : 'a matrix -> 'a array list;;
(** [to_arraylist m] returns the list of the arrays along the first direction
   of matrix [m]. It is equivalent to [[| m.(0); m.(1); ...;
   m.(]{!Unatural.to_int} [ni) |]] where [ni] is the first dimension of matrix
   [m]. *)

val of_arraylist : 'a array list -> 'a matrix;;
(** [of_arraylist al] returns a fresh matrix containing the array of arrays
   list [al].
   
   Raise [Invalid_argument], if all the arrays of list [al] do not have the
   same length. *)

val to_list : 'a matrix -> 'a list list;;
(** [to_list m] returns the list of lists of the elements of [m]. *)

val of_list : 'a list list -> 'a matrix;;
(** [of_list l] returns a fresh matrix containing the elements of the list of
   lists [l]. *)

(** {6 Iterators} *)

val iter : ('a -> unit) -> 'a matrix -> unit;;
(** [iter f m] applies function [f] in turn to all the elements of [m]. It is
   equivalent to [Array.iter f m.(0); Array.iter f m.(1); ...;
   Array.iter f m.(]{!Unatural.to_int} [ni - 1); ()] where [ni] is the first
   dimension of matrix [m]. *)

val map : ('a -> 'b) -> 'a matrix -> 'b matrix;;
(** [map f m] applies function [f] to all the elements of [m], and builds a
   matrix with the results returned by [f]: [[| Array.map f m.(0);
   Array.map f m.(1); ...; Array.map f a.(]{!Unatural.to_int} [ni - 1) |]]
   where [ni] is the first dimension of matrix [m]. *)

val iter_ij : (int -> int -> 'a -> unit) -> 'a matrix -> unit;;
(** Same as {!iter}, but the function is applied to the indices of the element
   as first and second arguments, and the element itself as third argument. *)

val map_ij : (int -> int -> 'a -> 'b) -> 'a matrix -> 'b matrix;;
(** Same as {!map}, but the function is applied to the indices of the element
   as first and second arguments, and the element itself as third argument. *)

val iter_i : ('a array -> unit) -> 'a matrix -> unit;;
(** [iter_i f m] applies function [f] in turn to all the rows of [m]. It is
   equivalent to [Array.iter f m]. *)

val iter_j : ('a array -> unit) -> 'a matrix -> unit;;
(** Same as {!iter_i}, but applied to the columns of the matrix argument. It
   is equivalent to {!iter_i} [f (]{!transpose} [m)]. *)

val map_i : ('a array -> 'b array) -> 'a matrix -> 'b matrix;;
(** [map_i f m] applies function [f] in turn to all the rows of [m]. It is
   equivalent to [Array.map f m]. *)

val map_j : ('a array -> 'b array) -> 'a matrix -> 'b matrix;;
(** Same as {!map_i}, but applied to the columns of the matrix. It is
   equivalent to {!transpose} [(]{!map_i} [f (]{!transpose} [m))]. *)

val fold_left : ('a -> 'b -> 'a) -> 'a -> 'b matrix -> 'a;;
(** [fold_left f x m] computes [Array.fold_left f (... (Array.fold_left f
   (Array.fold_left f x m.(0)) m.(1)) ...) m.(]{!Unatural.to_int} [ni - 1)],
   where [ni] is the first dimension of matrix [m]. *)

val fold_right : ('a -> 'b -> 'b) -> 'a matrix -> 'b -> 'b;;
(** [fold_right f m x] computes [Array.fold_right f m.(0)
   (Array.fold_right f m.(1) (...
   (Array.fold_right f m.(]{!Unatural.to_int} [ni - 1) x) ...))], where [ni]
   is the first dimension of matrix [m]. *)

val sub_iter :
  ('a -> unit) -> 'a matrix -> int -> Unatural.t -> int -> Unatural.t -> unit
;;
(** [sub_iter f m oi li oj lj] applies function [f] in turn to the elements
   number [(oi, oj)] to
   [(oi +] {!Unatural.to_int} [li - 1, oj +] {!Unatural.to_int} [lj - 1)] of
   matrix [m]. It is equivalent to {!Uarray.sub_iter} [f m.(oi) oj lj;]
   {!Uarray.sub_iter} [f m.(oi + 1) oj lj; ...;]
   {!Uarray.sub_iter} [f m.(oi +] {!Unatural.to_int} [li - 1) oj lj; ()]. *)

val sub_iter_ij :
  (int -> int -> 'a -> unit) -> 'a matrix ->
  int -> Unatural.t -> int -> Unatural.t -> unit
;;
(** Same as {!sub_iter}, but the function is applied to the indices of the
   element as first and second arguments, and the element itself as third
   argument. *)

(** {6 Iterators on two matrices} *)

val iter2 : ('a -> 'b -> unit) -> 'a matrix -> 'b matrix -> unit;;
(** [iter2 f m1 m2] applies function [f] in turn to all the elements of [m1]
   and [m2]. It is equivalent to {!Uarray.iter2} [f m1.(0) m2.(0);]
   {!Uarray.iter2} [f m1.(1) m2.(1); ...;]
   {!Uarray.iter2} [f m1.(]{!Unatural.to_int} [ni - 1)
   m2.(]{!Unatural.to_int} [ni - 1); ()], where [ni] is the first dimension of
   matrices [m1] and [m2]. *)

val map2 : ('a -> 'b -> 'c) -> 'a matrix -> 'b matrix -> 'c matrix;;
(** [map2 f m1 m2] applies function [f] to all the elements of [m1] and [m2],
   and builds a matrix with the results returned by [f]:
   [[|] {!Uarray.map2} [f m1.(0) m2.(0);] {!Uarray.map2} [f m1.(1) m2.(1);
   ...;] {!Uarray.map2} [f m1.(]{!Unatural.to_int} [ni - 1)
   m2.(]{!Unatural.to_int} [ni - 1) |]], where [ni] is the first dimension of
   matrices [m1] and [m2]. *)

val iter_ij2 :
  (int -> int -> 'a -> 'b -> unit) -> 'a matrix -> 'b matrix -> unit
;;
(** Same as {!iter2}, but the function is applied to the indices of the
   elements as first and second arguments, and the elements themselves as
   third and fourth arguments. *)

val map_ij2 :
  (int -> int -> 'a -> 'b -> 'c) -> 'a matrix -> 'b matrix -> 'c matrix
;;
(** Same as {!map2}, but the function is applied to the indices of the
   elements as first and second arguments, and the elements themselves as
   third and fourth arguments. *)

val fold_left2 :
  ('a -> 'b -> 'c -> 'a) -> 'a -> 'b matrix -> 'c matrix -> 'a
;;
(** [fold_left2 f x m1 m2] computes
   {!Uarray.fold_left2} [f (... (]{!Uarray.fold_left2} [f
   (]{!Uarray.fold_left2} [f x m1.(0) m2.(0)) m1.(1) m2.(1)) ...)
   m1.(]{!Unatural.to_int} [ni - 1) m2.(]{!Unatural.to_int} [ni - 1)], where
   [ni] is the first dimension of the matrices [m1] and [m2]. *)

val fold_right2 :
  ('a -> 'b -> 'c -> 'c) -> 'a matrix -> 'b matrix -> 'c -> 'c
;;
(** [fold_right2 f m1 m2 x] computes {!Uarray.fold_right2} [f m1.(0) m2.(0)
   (]{!Uarray.fold_right2} [f m1.(1) m2.(1) (...
   (]{!Uarray.fold_right2} [f m1.(]{!Unatural.to_int} [ni - 1)
    m2.(]{!Unatural.to_int} [ni - 1) x) ...))], where [ni] is the first
    dimension of matrices [m1] and [m2]. *)

(** {6 Specific applications} *)

val make_diagonal_matrix : 'a -> 'a array -> 'a matrix;;
(** [make_diagonal_matrix zero v] returns a fresh diagonal matrix with its main
   diagonal filled with elements of the vector [v] and all other entries filled
   with [zero]. *)

val make_int_diagonal_matrix : int array -> int matrix;;
val make_float_diagonal_matrix : float array -> float matrix;;
(** Specializations of {!make_diagonal_matrix} for integers and floats with the
   canonical zero value. *)

val resize : Unatural.t -> Unatural.t -> 'a matrix -> 'a matrix;;
(** [resize ni nj m] returns a fresh matrix of size [(ni, nj)]. Each element
   of the given matrix [m] becomes a submatrix of size ...
   
   Raise [Invalid_arg] if the matrix [m] has a zero dimension and not ... . *)

val zoom : Unatural.t -> Unatural.t -> 'a matrix -> 'a matrix;;
(** [zoom ni nj m] returns a fresh matrix whose dimensions are those of [m]
   multiplied by [ni] and [nj]. Each element of the given matrix [m] becomes
   a submatrix of size [(ni, nj)]. *)

(**/**)
(** {6 Undocumented functions} *)

external unsafe_of_array_array : 'a array array -> 'a matrix = "%identity";;

val unsafe_get : 'a matrix -> int -> int -> 'a;;
val unsafe_set : 'a matrix -> int -> int -> 'a -> unit;;
