(***********************************************************************)
(*                                                                     *)
(*                                Ulib                                 *)
(*                                                                     *)
(*                   OCaml user's additional library                   *)
(*                                                                     *)
(*               EPI Pomdapi, INRIA Paris - Rocquencourt               *)
(*                                                                     *)
(*  Copyright 2004-2014 INRIA.                                         *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: uset.ml,v 1.3 2014/01/10 15:28:04 fclement Exp $ *)

let s_invalid_arg = Upervasives.invalid_arg "Uset";;

type 'a uindex = 'a Unonempty_set.uindex
and 'a index = 'a Unonempty_set.index
;;

type 'a t =
  | Empty
  | Nonempty of 'a Unonempty_set.t
;;

(* Construction functions. *)

let empty () = Empty;;

let singleton u uindex = Nonempty (Unonempty_set.singleton u uindex);;

let full u = Nonempty (Unonempty_set.full u);;

let of_array u uindices =
  let len = Uarray.length uindices in
  if len = Unatural._0_
  then Empty
  else Nonempty (Unonempty_set.of_array u uindices)
;;

let of_list u = function
  | [] -> Empty
  | x :: l ->
      let uindices = (x, l) in
      Nonempty (Unonempty_set.of_list u uindices)
;;

let of_rectangle h w = Nonempty (Unonempty_set.of_rectangle h w);;

let of_natural_rectangle h w =
  Nonempty (Unonempty_set.of_natural_rectangle h w)
;;

(* Modification functions. *)

let add u s uindex =
  match s with
  | Empty -> singleton u uindex
  | Nonempty ne -> Unonempty_set.add ne uindex; s
;;

(* Indices and elements. *)

let cardinal = function
  | Empty -> Unatural._0_
  | Nonempty ne -> Unonempty_set.cardinal ne
;;

let valid_index s index =
  match s with
  | Empty -> false
  | Nonempty ne -> Unonempty_set.valid_index ne index
;;

let check_index fun_name s index =
  if not (valid_index s index) then
    s_invalid_arg fun_name
      (Printf.sprintf "index %i out of set of %d element(s)"
         (Uindex.to_int index)
         (Unatural.to_int (cardinal s)))
;;

let mem s uindex =
  match s with
  | Empty -> false
  | Nonempty ne -> Unonempty_set.mem ne uindex
;;

let choice = function
  | Empty -> s_invalid_arg "choice" "empty set"
  | Nonempty ne -> Unonempty_set.choice ne
;;

let uindex s index =
  match s with
  | Empty -> s_invalid_arg "indice" "empty set"
  | Nonempty ne -> Unonempty_set.uindex ne index
;;

let uindices = function
  | Empty -> [||]
  | Nonempty ne -> Unonempty_set.uindices ne
;;

let element s index =
  match s with
  | Empty -> s_invalid_arg "element" "empty set"
  | Nonempty ne -> Unonempty_set.element ne index
;;

let elements = function
  | Empty -> [||]
  | Nonempty ne -> Unonempty_set.elements ne
;;

(* Iterators. *)

let fold_left f accu = function
  | Empty -> accu
  | Nonempty ne -> Unonempty_set.fold_left f accu ne
;;

let fold_left_i f accu = function
  | Empty -> accu
  | Nonempty ne -> Unonempty_set.fold_left_i f accu ne
;;

let union s1 s2 =
  match (s1, s2) with
  | (Empty, s) | (s, Empty) -> s
  | (Nonempty ne1, Nonempty ne2) -> Nonempty (Unonempty_set.union ne1 ne2)
;;

let iter f = function
  | Empty -> ()
  | Nonempty ne -> Unonempty_set.iter f ne
;;

let iter_i f = function
  | Empty -> ()
  | Nonempty ne -> Unonempty_set.iter_i f ne
;;

let map_indices f = function
  | Empty -> [||]
  | Nonempty ne -> Unonempty_set.map_indices f ne
;;

let map f s u =
  match s with
  | Empty -> Empty
  | Nonempty ne -> Nonempty (Unonempty_set.map f ne u)
;;

let find p = function
  | Empty -> raise Not_found
  | Nonempty ne -> Unonempty_set.find p ne
;;

let find_all p = function
  | Empty -> raise Not_found
  | Nonempty ne -> Unonempty_set.find_all p ne
;;

let filter = find_all;;

(* Subsets. *)

type 'a subset =
  {
   support : 'a t;
   indices : 'a index array;
   set : 'a t;
  }
;;

let full_index_of_uindex = Unonempty_set.full_index_of_uindex;;

let subset_of_nonempty_subset
    { Unonempty_set.support = sup_ne;
      indices = indices;
      nonempty_set = ne; } =
  { support = Nonempty sup_ne; indices = indices; set = Nonempty ne; }
;;

(* We use here the fact that the full set, containing all the elements of the
  universe, is numbered in the same order as the universe.
  Hence, the use of magic function index_of_uindex.
  If efficiency is an issue, consider using Obj.magic on the whole array of
  indices. *)
let subset_of_full full s =
  match (full, s) with
  | (Empty, _) -> assert false
  | (Nonempty _, Empty) -> { support = full; indices = [||]; set = s; }
  | (Nonempty full_ne, Nonempty ne) ->
      subset_of_nonempty_subset (Unonempty_set.subset_of_full full_ne ne)
;;

type 'a dichotomy = 'a subset * 'a subset;;

let dichotomy p = function
  | Empty as s ->
      let empty_ss = { support = s; indices = [||]; set = s; } in
      (empty_ss, empty_ss)
  | Nonempty ne as s ->
      let dicho (tl, fl) index uindex =
        if p index uindex then (index :: tl, fl) else (tl, index :: fl) in
      let (tl, fl) = fold_left_i dicho ([], []) s in
      let uindex = Unonempty_set.uindex ne in
      let (utl, ufl) = (Ulist.rev_map uindex tl, Ulist.rev_map uindex fl) in
      let (ts, fs) =
        (of_list ne.Unonempty_set.universe utl,
         of_list ne.Unonempty_set.universe ufl) in
      ({ support = s; indices = Array.of_list tl; set = ts; },
       { support = s; indices = Array.of_list fl; set = fs; })
;;
